# used by paper: TSB-UAD as the main evaluator
# github: https://github.com/johnpaparrizos/TSB-UAD/blob/main/TSB_AD/utils/metrics.py
import numpy as np
from sklearn import metrics
from metrics.evaluate_utils import find_length,range_convers_new


def extend_postive_range(x, window=16):
    label = x.copy().astype(float)
#     print(label)
    L = range_convers_new(label)  # index of non-zero segments
#     print(L)
    length = len(label)
    for k in range(len(L)):
        s = L[k][0]
        e = L[k][1]
        # x1 is the extended list like [1,2,3] which are non-zero(from the end-e)
        x1 = np.arange(e, min(e + window // 2, length))
        label[x1] += np.sqrt(1 - (x1 - e) / (window))
        # before the start-s
        x2 = np.arange(max(s - window // 2, 0), s)
        label[x2] += np.sqrt(1 - (s - x2) / (window))

    label = np.minimum(np.ones(length), label)
    return label


def extend_postive_range_individual(x, percentage=0.2):
    label = x.copy().astype(float)
    L = range_convers_new(label)  # index of non-zero segments
    length = len(label)
    for k in range(len(L)):
        s = L[k][0]
        e = L[k][1]

        l0 = int((e - s + 1) * percentage)

        x1 = np.arange(e, min(e + l0, length))
        label[x1] += np.sqrt(1 - (x1 - e) / (2 * l0))

        x2 = np.arange(max(s - l0, 0), s)
        label[x2] += np.sqrt(1 - (s - x2) / (2 * l0))

    label = np.minimum(np.ones(length), label)
    return label


def TPR_FPR_RangeAUC(labels, pred, P, L):
    product = labels * pred

    TP = np.sum(product)

    # recall = min(TP/P,1)
    P_new = (P + np.sum(labels)) / 2  # so TPR is neither large nor small
    # P_new = np.sum(labels)
    recall = min(TP / P_new, 1)
    # recall = TP/np.sum(labels)
    # print('recall '+str(recall))

    existence = 0
    for seg in L:
        if np.sum(product[seg[0]:(seg[1] + 1)]) > 0:
            existence += 1

    existence_ratio = existence / len(L)
    # print(existence_ratio)

    # TPR_RangeAUC = np.sqrt(recall*existence_ratio)
    # print(existence_ratio)
    TPR_RangeAUC = recall * existence_ratio

    FP = np.sum(pred) - TP
    # TN = np.sum((1-pred) * (1-labels))

    # FPR_RangeAUC = FP/(FP+TN)
    N_new = len(labels) - P_new
    FPR_RangeAUC = FP / N_new

    Precision_RangeAUC = TP / np.sum(pred)

    return TPR_RangeAUC, FPR_RangeAUC, Precision_RangeAUC


def Range_AUC(score_t_test, y_test,  window=5, percentage=0, plot_ROC=False, AUC_type='window'):
    # AUC_type='window'/'percentage'
    score = score_t_test
    labels = y_test
    score_sorted = -np.sort(-score)

    P = np.sum(labels)
    # print(np.sum(labels))
    if AUC_type == 'window':
        labels = extend_postive_range(labels, window=window)
    else:
        labels = extend_postive_range_individual(labels, percentage=percentage)

    # print(np.sum(labels))
    L = range_convers_new(labels)
    TPR_list = [0]
    FPR_list = [0]
    Precision_list = [1]

    for i in np.linspace(0, len(score) - 1, 250).astype(int):
        threshold = score_sorted[i]
        # print('thre='+str(threshold))
        pred = score >= threshold
        TPR, FPR, Precision = TPR_FPR_RangeAUC(labels, pred, P, L)

        TPR_list.append(TPR)
        FPR_list.append(FPR)
        Precision_list.append(Precision)

    TPR_list.append(1)
    FPR_list.append(1)  # otherwise, range-AUC will stop earlier than (1,1)

    tpr = np.array(TPR_list)
    fpr = np.array(FPR_list)
    prec = np.array(Precision_list)

    width = fpr[1:] - fpr[:-1]
    height = (tpr[1:] + tpr[:-1]) / 2
    AUC_range = np.sum(width * height)

    width_PR = tpr[1:-1] - tpr[:-2]
    height_PR = (prec[1:] + prec[:-1]) / 2
    AP_range = np.sum(width_PR * height_PR)

    if plot_ROC:
        return AUC_range, AP_range, fpr, tpr, prec

    return AUC_range


def point_wise_AUC(score_t_test, y_test,  plot_ROC=False):
    # area under curve
    label = y_test
    score = score_t_test
    auc = metrics.roc_auc_score(label, score)
    # plor ROC curve
    if plot_ROC:
        fpr, tpr, thresholds = metrics.roc_curve(label, score)
        # display = metrics.RocCurveDisplay(fpr=fpr, tpr=tpr, roc_auc=auc)
        # display.plot()
        return auc, fpr, tpr
    else:
        return auc


def main():
    y_test = np.zeros(100)
    y_test[10:20] = 1
    y_test[50:60] = 1
    pred_labels = np.zeros(100)
    pred_labels[15:17] = 0.5
    pred_labels[55:62] = 0.7
    # pred_labels[51:55] = 1
    # true_events = get_events(y_test)
    point_auc = point_wise_AUC(pred_labels, y_test)
    range_auc = Range_AUC(pred_labels, y_test)
    print("point_auc: {}, range_auc: {}".format(point_auc, range_auc))


if __name__ == "__main__":
    main()


import numpy as np
from statsmodels.tsa.stattools import acf
from scipy.signal import argrelextrema


def get_composite_fscore_from_scores(score_t_test, thres, true_events, prec_t, return_prec_rec=False):
    pred_labels = score_t_test > thres
    tp = np.sum([pred_labels[start:end + 1].any() for start, end in true_events.values()])
    fn = len(true_events) - tp
    rec_e = tp / (tp + fn)
    fscore_c = 2 * rec_e * prec_t / (rec_e + prec_t)
    if prec_t == 0 and rec_e == 0:
        fscore_c = 0
    if return_prec_rec:
        return prec_t, rec_e, fscore_c
    return fscore_c


class NptConfig:
    def __init__(self, config_dict):
        for k, v in config_dict.items():
            setattr(self, k, v)

def find_length(data):
    if len(data.shape) > 1:
        return 0
    data = data[:min(20000, len(data))]

    base = 3
    auto_corr = acf(data, nlags=400, fft=True)[base:]

    local_max = argrelextrema(auto_corr, np.greater)[0]
    try:
        max_local_max = np.argmax([auto_corr[lcm] for lcm in local_max])
        if local_max[max_local_max] < 3 or local_max[max_local_max] > 300:
            return 125
        return local_max[max_local_max] + base
    except:
        return 125


def range_convers_new(label):
    '''
    input: arrays of binary values
    output: list of ordered pair [[a0,b0], [a1,b1]... ] of the inputs
    '''
    L = []
    i = 0
    j = 0
    while j < len(label):
        while label[i] == 0:
            i += 1
            if i >= len(label):
                break
        j = i + 1
        if j >= len(label):
            if j == len(label):
                L.append((i, j - 1))
            break
        while label[j] != 0:
            j += 1
            if j >= len(label):
                L.append((i, j - 1))
                break
        if j >= len(label):
            break
        L.append((i, j - 1))
        i = j
    return L


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
测试CCE配置系统的脚本
"""

import os
import sys
import tempfile
import shutil
from pathlib import Path

# 添加src目录到路径
sys.path.insert(0, str(Path(__file__).parent / 'src'))

def test_config_system():
    """测试配置系统的基本功能"""
    print("🧪 测试CCE配置系统...")
    
    try:
        from cce.config import CCEConfig, get_config, get_datasets_path, create_user_config
        
        # 测试1: 创建配置实例
        print("\n1. 测试配置实例创建...")
        config = CCEConfig()
        print(f"   ✅ 配置实例创建成功")
        
        # 测试2: 获取数据集路径
        print("\n2. 测试数据集路径获取...")
        datasets_path = get_datasets_path()
        print(f"   ✅ 数据集路径: {datasets_path}")
        
        # 测试3: 创建用户配置文件
        print("\n3. 测试用户配置文件创建...")
        with tempfile.TemporaryDirectory() as temp_dir:
            config_path = Path(temp_dir) / 'test_config.yaml'
            created_path = create_user_config(str(config_path))
            print(f"   ✅ 配置文件已创建: {created_path}")
            
            # 验证配置文件内容
            if config_path.exists():
                print(f"   ✅ 配置文件存在")
                with open(config_path, 'r') as f:
                    content = f.read()
                    if 'datasets_path' in content:
                        print(f"   ✅ 配置文件包含数据集路径配置")
                    else:
                        print(f"   ❌ 配置文件缺少数据集路径配置")
            else:
                print(f"   ❌ 配置文件不存在")
        
        # 测试4: 测试配置优先级
        print("\n4. 测试配置优先级...")
        with tempfile.TemporaryDirectory() as temp_dir:
            # 创建测试配置文件
            test_config_path = Path(temp_dir) / 'cce_config.yaml'
            with open(test_config_path, 'w') as f:
                f.write("datasets_path: /test/datasets\n")
            
            # 切换到测试目录
            original_cwd = os.getcwd()
            os.chdir(temp_dir)
            
            try:
                # 创建新的配置实例（应该读取当前目录的配置文件）
                test_config = CCEConfig()
                test_datasets_path = test_config.get_datasets_path()
                print(f"   ✅ 当前目录配置优先级测试: {test_datasets_path}")
            finally:
                os.chdir(original_cwd)
        
        print("\n🎉 所有测试通过！")
        return True
        
    except Exception as e:
        print(f"\n❌ 测试失败: {e}")
        import traceback
        traceback.print_exc()
        return False

def test_cli_commands():
    """测试CLI配置命令"""
    print("\n🧪 测试CLI配置命令...")
    
    try:
        # 模拟CLI参数
        class MockArgs:
            def __init__(self, config_action, path=None):
                self.config_action = config_action
                self.path = path
        
        from cce.cli import handle_config_command
        
        # 测试show命令
        print("\n1. 测试config show命令...")
        args = MockArgs('show')
        handle_config_command(args)
        print("   ✅ config show命令执行成功")
        
        # 测试create命令
        print("\n2. 测试config create命令...")
        with tempfile.TemporaryDirectory() as temp_dir:
            config_path = Path(temp_dir) / 'test_config.yaml'
            args = MockArgs('create', str(config_path))
            handle_config_command(args)
            print("   ✅ config create命令执行成功")
        
        print("\n🎉 CLI测试通过！")
        return True
        
    except Exception as e:
        print(f"\n❌ CLI测试失败: {e}")
        import traceback
        traceback.print_exc()
        return False

if __name__ == '__main__':
    print("🚀 开始测试CCE配置系统...")
    
    success1 = test_config_system()
    success2 = test_cli_commands()
    
    if success1 and success2:
        print("\n🎉 所有测试都通过了！配置系统工作正常。")
        sys.exit(0)
    else:
        print("\n❌ 部分测试失败，请检查配置系统。")
        sys.exit(1)


"""
This function is adapted from [M2N2] by [Dongmin Kim et al.]
Original source: [https://github.com/carrtesy/M2N2]
Reimplemented by: [EmorZz1G]
"""


import os
import pandas as pd
from tqdm import tqdm
import matplotlib.pyplot as plt
import pickle
import math

# Trainer

class Trainer:
    '''
    Prepares Offline-trained models.
    '''
    def __init__(self, model, train_loader):
        self.model = model
        self.train_loader = train_loader


    def train(self):
        raise NotImplementedError()

    @DeprecationWarning
    def checkpoint(self, filepath):
        self.logger.info(f"checkpointing: {filepath} @Trainer - torch.save")
        torch.save(self.model.state_dict(), filepath)

    @DeprecationWarning
    def load(self, filepath):
        self.logger.info(f"loading: {filepath} @Trainer - torch.load_state_dict")
        self.model.load_state_dict(torch.load(filepath))
        self.model.to(self.args.device)


    @staticmethod
    def save_dictionary(dictionary, filepath):
        with open(filepath, "wb") as f:
            pickle.dump(dictionary, f)
            
import torch
import numpy as np

import matplotlib
import matplotlib.pyplot as plt

import pickle
import os

import pandas as pd

import json
from ast import literal_eval



# matplotlib.rcParams['agg.path.chunksize'] = 10000


class Tester:
    '''
    Test-time logics,
    including offline evaluation and online adaptation.
    '''
    def __init__(self, model, train_loader, test_loader):
        self.model = model
        self.train_loader = train_loader
        self.test_loader = test_loader


    def calculate_anomaly_scores(self, dataloader):
        raise NotImplementedError()


    def checkpoint(self, filepath):
        self.logger.info(f"checkpointing: {filepath} @Trainer - torch.save")
        torch.save(self.model.state_dict(), filepath)


    def load(self, filepath):
        self.logger.info(f"loading: {filepath} @Trainer - torch.load_state_dict")
        self.model.load_state_dict(torch.load(filepath))
        self.model.to(self.args.device)


    def load_trained_model(self):
        self.load(os.path.join(self.args.checkpoint_path, f"best.pth"))


    @staticmethod
    def save_dictionary(dictionary, filepath):
        with open(filepath, "wb") as f:
            pickle.dump(dictionary, f)


    def prepare_stats(self):
        '''
        prepare anomaly scores of train data / test data.
        '''
        # train
        train_anoscs_pt_path = os.path.join(self.args.output_path, "train_anoscs.pt")
        if self.args.load_anoscs and os.path.isfile(train_anoscs_pt_path):
            self.logger.info("train_anoscs.pt file exists, loading...")
            with open(train_anoscs_pt_path, 'rb') as f:
                train_anoscs = torch.load(f)
                train_anoscs.to(self.args.device)
            self.logger.info(f"{train_anoscs.shape}")
        else:
            self.logger.info("train_anoscs.pt file does not exist, calculating...")
            train_anoscs = torch.Tensor(self.calculate_anomaly_scores(self.train_loader))  # (B, L, C) => (B, L)
            self.logger.info("saving train_anoscs.pt...")
            with open(train_anoscs_pt_path, 'wb') as f:
                torch.save(train_anoscs, f)
        torch.cuda.empty_cache()

        # test
        test_anosc_pt_path = os.path.join(self.args.output_path, "test_anoscs.pt")
        if self.args.load_anoscs and os.path.isfile(test_anosc_pt_path):
            self.logger.info("test_anoscs.pt file exists, loading...")
            with open(test_anosc_pt_path, 'rb') as f:
                test_anoscs = torch.load(f)
                test_anoscs.to(self.args.device)
            self.logger.info(f"{test_anoscs.shape}")
        else:
            self.logger.info("test_anoscs.pt file does not exist, calculating...")
            test_anoscs = torch.Tensor(self.calculate_anomaly_scores(self.test_loader))  # (B, L, C) => (B, L)
            self.logger.info("saving test_anoscs.pt...")
            with open(test_anosc_pt_path, 'wb') as f:
                torch.save(test_anoscs, f)
        torch.cuda.empty_cache()

        # train_anoscs, test anoscs (T=B*L, ) and ground truth
        train_mask = (self.train_loader.dataset.y != -1)
        self.train_anoscs = train_anoscs.detach().cpu().numpy()[train_mask] # does not include -1's
        self.test_anoscs = test_anoscs.detach().cpu().numpy() # may include -1's, filtered when calculating final results.
        self.gt = self.test_loader.dataset.y

        # thresholds for visualization
        self.th_q95 = np.quantile(self.train_anoscs, 0.95)
        self.th_q99 = np.quantile(self.train_anoscs, 0.99)
        self.th_q100 = np.quantile(self.train_anoscs, 1.00)
        # self.th_off_f1_best = get_best_static_threshold(gt=self.gt, anomaly_scores=self.test_anoscs)


    def infer(self, mode, cols):
        result_df = pd.DataFrame(columns=cols)
        gt = self.test_loader.dataset.y

        # for single inference: select specific threshold tau
        th = self.args.thresholding
        if th[0] == "q":
            th = float(th[1:]) / 100
            tau = np.quantile(self.train_anoscs, th)
        elif th == "off_f1_best":
            tau = self.th_off_f1_best
        else:
            raise ValueError(f"Thresholding mode {self.args.thresholding} is not supported.")

        # get result
        if mode == "offline":
            anoscs, pred = self.offline(tau)


        elif mode == "online":
            anoscs, pred = self.online(self.test_loader, tau, normalization=self.args.normalization)
            # result = get_summary_stats(gt, pred)
            # roc_auc = calculate_roc_auc(gt, anoscs,
            #                             path=self.args.output_path,
            #                             save_roc_curve=self.args.save_roc_curve,
            #                             drop_intermediate=False,
            #                             )
            # result["ROC_AUC"] = roc_auc

            # pr_auc = calculate_pr_auc(gt, anoscs,
            #                           path=self.args.output_path,
            #                           save_pr_curve=self.args.save_pr_curve,
            #                           )
            # result["PR_AUC"] = pr_auc

            # result_df = pd.DataFrame([result], index=[mode], columns=result_df.columns)
            # result_df.at[mode, "tau"] = tau


        return result_df


    def online(self, *args):
        raise NotImplementedError()

# models
'''
Basic MLP implementation by:
Dongmin Kim (tommy.dm.kim@gmail.com)
'''

import torch
import torch.nn as nn

import torch
import torch.nn as nn


class Detrender(nn.Module):
    def __init__(self, num_features: int, gamma=0.99):
        """
        :param num_features: the number of features or channels
        :param eps: a value added for numerical stability
        """
        super(Detrender, self).__init__()
        self.num_features = num_features
        self.gamma = gamma
        self.mean = nn.Parameter(torch.zeros(1, 1, self.num_features), requires_grad=False)


    def forward(self, x, mode:str):
        if mode == 'norm':
            x = self._normalize(x)
        elif mode == 'denorm':
            x = self._denormalize(x)
        else: raise NotImplementedError
        return x


    def _update_statistics(self, x):
        dim2reduce = tuple(range(0, x.ndim-1))
        mu = torch.mean(x, dim=dim2reduce, keepdim=True).detach()
        self.mean.lerp_(mu, 1-self.gamma)


    def _set_statistics(self, x:torch.Tensor):
        self.mean = nn.Parameter(x, requires_grad=False)


    def _normalize(self, x):
        x = x - self.mean
        return x


    def _denormalize(self, x):
        x = x + self.mean
        return x


class MLP(nn.Module):
    def __init__(self, seq_len, num_channels, latent_space_size, gamma, normalization="None"):
        super().__init__()
        self.L, self.C = seq_len, num_channels
        self.encoder = Encoder(seq_len*num_channels, latent_space_size)
        self.decoder = Decoder(seq_len*num_channels, latent_space_size)
        self.normalization = normalization

        if self.normalization == "Detrend":
            self.use_normalizer = True
            self.normalizer = Detrender(num_channels, gamma=gamma)
        else:
            self.use_normalizer = False


    def forward(self, X):
        B, L, C = X.shape
        assert (L == self.L) and (C == self.C)

        if self.use_normalizer:
            X = self.normalizer(X, "norm")
            
        z = self.encoder(X.reshape(B, L*C))
        out = self.decoder(z).reshape(B, L, C)

        if self.use_normalizer:
            out = self.normalizer(out, "denorm")
        return out


class Encoder(nn.Module):
    def __init__(self, input_size, latent_space_size):
        super().__init__()
        self.linear1 = nn.Linear(input_size, input_size // 2)
        self.relu1 = nn.ReLU()
        self.linear2 = nn.Linear(input_size // 2, input_size // 4)
        self.relu2 = nn.ReLU()
        self.linear3 = nn.Linear(input_size // 4, latent_space_size)
        self.relu3 = nn.ReLU()


    def forward(self, x):
        x = self.linear1(x)
        x = self.relu1(x)
        x = self.linear2(x)
        x = self.relu2(x)
        x = self.linear3(x)
        x = self.relu3(x)
        return x


class Decoder(nn.Module):
    def __init__(self, input_size, latent_space_size):
        super().__init__()
        self.linear1 = nn.Linear(latent_space_size, input_size // 4)
        self.relu1 = nn.ReLU()
        self.linear2 = nn.Linear(input_size // 4, input_size // 2)
        self.relu2 = nn.ReLU()
        self.linear3 = nn.Linear(input_size // 2, input_size)


    def forward(self, x):
        x = self.linear1(x)
        x = self.relu1(x)
        x = self.linear2(x)
        x = self.relu2(x)
        out = self.linear3(x)
        return out

# others
import torch
import torch.nn.functional as F
import numpy as np


class MLP_Trainer(Trainer):
    def __init__(self, model, train_loader, epochs=10, lr=1e-3, L2_reg=0, device='cuda'):
        super(MLP_Trainer, self).__init__(model, train_loader)

        self.model = model
        self.device = device
        self.epochs = epochs
        self.optimizer = torch.optim.AdamW(params=self.model.parameters(), lr=lr, weight_decay=L2_reg) # L2 Reg is set to zero by default, but can be set as needed.


    def train(self):

        train_iterator = tqdm(
            range(1, self.epochs + 1),
            total=self.epochs,
            desc="training epochs",
            leave=True
        )

        best_train_stats = None
        for epoch in train_iterator:
            train_stats = self.train_epoch()
            # self.logger.info(f"epoch {epoch} | train_stats: {train_stats}")
            # self.checkpoint(os.path.join(self.args.checkpoint_path, f"epoch{epoch}.pth"))

            # if best_train_stats is None or train_stats < best_train_stats:
            #     self.logger.info(f"Saving best results @epoch{epoch}")
            #     self.checkpoint(os.path.join(self.args.checkpoint_path, f"best.pth"))
            #     best_train_stats = train_stats


    def train_epoch(self):
        self.model.train()
        train_summary = 0.0
        for i, batch_data in enumerate(self.train_loader):
            train_log = self._process_batch(batch_data)
            train_summary += train_log["summary"]
        train_summary /= len(self.train_loader)
        return train_summary


    def _process_batch(self, batch_data) -> dict:
        X = batch_data[0].to(self.device)
        B, L, C = X.shape

        # recon
        Xhat = self.model(X)

        # optimize
        self.optimizer.zero_grad()
        loss = F.mse_loss(Xhat, X)
        loss.backward()
        self.optimizer.step()

        out = {
            "recon_loss": loss.item(),
            "summary": loss.item(),
        }
        return out


class MLP_Tester(Tester):
    def __init__(self, model, train_loader, test_loader, device='cuda', load=False, lr=1e-3):
        super(MLP_Tester, self).__init__(model, train_loader, test_loader)

        self.model = model
        self.device = device
        self.lr = lr

        if load:
            self.load_trained_model()
            self.prepare_stats()


    @torch.no_grad()
    def calculate_anomaly_scores(self, dataloader):
        recon_errors = self.calculate_recon_errors(dataloader) # B, L, C
        #anomaly_scores = recon_errors.mean(dim=2).reshape(-1).detach().cpu() # B, L -> (T=B*L, )
        anomaly_scores = recon_errors.mean(axis=2).reshape(-1)
        return anomaly_scores


    @torch.no_grad()
    def calculate_recon_errors(self, dataloader):
        '''
        :return:  returns (B, L, C) recon loss tensor
        '''
        it = tqdm(
            dataloader,
            total=len(dataloader),
            desc="calculating reconstruction errors",
            leave=True
        )
        recon_errors = []
        Xs, Xhats = [], []
        for i, batch_data in enumerate(it):
            X = batch_data[0].to(self.device)
            B, L, C = X.shape
            Xhat = self.model(X)

            recon_error = F.mse_loss(Xhat, X, reduction='none')
            recon_error = recon_error.detach().cpu().numpy()
            recon_errors.append(recon_error)
            torch.cuda.empty_cache()

        recon_errors = np.concatenate(recon_errors, axis=0)
        return recon_errors


    def online(self, dataloader, init_thr, normalization="None"):
        # self.load_trained_model() # reset

        it = tqdm(
            dataloader,
            total=len(dataloader),
            desc="inference",
            leave=True
        )

        tau = init_thr
        TT_optimizer = torch.optim.SGD([p for p in self.model.parameters()], lr=self.lr)

        Xs, Xhats = [], []
        preds = []
        As, thrs = [], []

        for i, batch_data in enumerate(it):
            X = batch_data[0].to(self.device)
            B, L, C = X.shape

            # Update of test-time statistics.
            if normalization == "Detrend":
                self.model.normalizer._update_statistics(X)

            # inference
            Xhat = self.model(X)
            E = (Xhat-X)**2
            A = E.mean(dim=2)
            # A: (B, L, C) -> (B, L)
            ytilde = (A >= tau).float()
            pred = ytilde

            # log model outputs
            Xs.append(X)
            Xhats.append(Xhat.clone().detach())
            As.append(A[:,-1].clone().detach())
            preds.append(pred.clone().detach())
            thrs.append(tau)

            # learn new-normals
            TT_optimizer.zero_grad()
            mask = (ytilde == 0)
            recon_loss = (A * mask).mean()
            recon_loss.backward()
            TT_optimizer.step()


        # outputs
        anoscs = torch.cat(As, axis=0).reshape(-1).detach().cpu().numpy()

        return anoscs
    
    

from .base import BaseDetector
from ..utils.torch_utility import EarlyStoppingTorch, get_gpu
from torch.utils.data import DataLoader
from ..utils.dataset import ReconstructDataset

class M2N2(BaseDetector):
    def __init__(self, 
                 win_size=12,
                 stride=12,
                 num_channels=1, 
                 batch_size=64,
                 epochs=10,
                 latent_dim=128,
                 lr=1e-3,
                 normalization="Detrend",
                 gamma=0.99,
                 th='q95'):

        self.model_name = 'M2N2'
        self.normalization = normalization
        self.device = get_gpu(True)
        self.model = MLP(
            seq_len=win_size,
            num_channels=num_channels,
            latent_space_size=latent_dim,
            gamma=gamma,
            normalization=normalization,
        ).to(self.device)
        
        self.th = float(th[1:]) / 100
        self.lr = lr
        self.epochs = epochs
        self.batch_size = batch_size
        self.win_size = win_size
        self.stride = stride
        self.validation_size = 0.1
        
        
        
        
    def fit(self, data):
        print("======================TRAIN MODE======================")

        tsTrain = data[:int((1-self.validation_size)*len(data))]
        tsValid = data[int((1-self.validation_size)*len(data)):]

        train_loader = DataLoader(
            dataset=ReconstructDataset(tsTrain, window_size=self.win_size, stride=self.stride),
            batch_size=self.batch_size,
            shuffle=True
        )
        
        self.trainer = MLP_Trainer(
            model=self.model,
            train_loader=train_loader,
            epochs=self.epochs,
            lr=self.lr,
            device=self.device
        )
        
        self.trainer.train()
        
        self.tester = MLP_Tester(
            model=self.model,
            train_loader=train_loader,
            test_loader=train_loader,
            device=self.device,
            load=False,
            lr=self.lr
        )
        self.model.eval()
        
        train_anoscs = self.tester.calculate_anomaly_scores(train_loader)
        
        self.tau = np.quantile(train_anoscs, self.th)
        
    
    def decision_function(self, data):
        self.model.eval()

        print("======================TEST MODE======================")

        
        test_loader = DataLoader(
            dataset=ReconstructDataset(data, window_size=self.win_size, stride=self.stride),
            batch_size=self.batch_size,
            shuffle=False,
        )
        
        
        tester = MLP_Tester(
            model=self.model,
            train_loader=test_loader,
            test_loader=test_loader,
            device=self.device,
            load=False,
            lr=self.lr
        )
        
        anoscs = tester.online(test_loader, self.tau, normalization=self.normalization)
        
        self.__anomaly_score = anoscs

        if self.__anomaly_score.shape[0] < len(data):
            self.__anomaly_score = np.array([self.__anomaly_score[0]]*math.ceil((self.win_size-1)/2) + 
                        list(self.__anomaly_score) + [self.__anomaly_score[-1]]*((self.win_size-1)//2))
        
        return anoscs
        
        
        
        


import numpy as np
import logging
import math
from stumpy import stumpi
from TSB_AD.models.base import BaseDetector
from TSB_AD.utils.utility import zscore

class Left_STAMPi(BaseDetector):

    def __init__(self, n_init_train=100, window_size=50, normalize=True):
        super().__init__()
        self.n_init_train = n_init_train
        self.window_size = window_size
        self.normalize = normalize

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.

        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        n_samples, n_features = X.shape
        if self.normalize: 
            X = zscore(X, axis=0, ddof=0)

        warmup = self.n_init_train
        ws = self.window_size

        if ws > warmup:
            logging.warning(f"WARN: window_size is larger than n_init_train. Adjusting to n_init_train={warmup}.")
            ws = warmup
        if ws < 3:
            logging.warning("WARN: window_size must be at least 3. Adjusting to 3.")
            ws = 3

        self.stream = stumpi(X[:warmup, 0], m=ws, egress=False)
        for point in X[warmup:, 0]:
            self.stream.update(point)
  
        self.decision_scores_ = self.stream.left_P_
        self.decision_scores_[:warmup] = 0  

        return self

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.

        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples.

        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        n_samples = X.shape[0]
        padded_scores = self.pad_anomaly_scores(self.decision_scores_, n_samples, self.window_size)
        return padded_scores

    @staticmethod
    def pad_anomaly_scores(scores, n_samples, window_size):
        """
        Pads the anomaly scores to match the length of the input time series.
        Padding is symmetric, using the first and last values.
        """
        left_padding = [scores[0]] * math.ceil((window_size - 1) / 2)
        right_padding = [scores[-1]] * ((window_size - 1) // 2)
        padded_scores = np.array(left_padding + list(scores) + right_padding)

        return padded_scores[:n_samples]


"""
Evaluation module for CCE (Confidence-Consistency Evaluation).

This module contains evaluation metrics and analysis tools for time series anomaly detection.
"""
from . import eval_metrics
from . import analysis_metrics

__all__ = [
    'eval_metrics',
    'analysis_metrics',
]


from sklearn.metrics import confusion_matrix
import numpy as np


def MCC(y_test, pred_labels):
    tn, fp, fn, tp = confusion_matrix(y_test, pred_labels).ravel()
    MCC_score = (tp*tn-fp*fn)/(((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn))**0.5)

    return MCC_score


def main():
    y_test = np.zeros(100)
    y_test[10:20] = 1
    y_test[50:60] = 1
    pred_labels = np.zeros(100)
    pred_labels[15:17] = 1
    pred_labels[55:62] = 1
    # pred_labels[51:55] = 1
    # true_events = get_events(y_test)
    confusion_matric = MCC(y_test, pred_labels)
#     print(confusion_matric)


if __name__ == "__main__":
    main()



# used by paper: Exathlon: A Benchmark for Explainable Anomaly Detection over Time Series_VLDB 2021
# github: https://github.com/exathlonbenchmark/exathlon
import numpy as np
from metrics.evaluate_utils import range_convers_new

# the existence reward on the bias
def b(bias, i, length):
    if bias == 'flat':
        return 1
    elif bias == 'front-end bias':
        return length - i + 1
    elif bias == 'back-end bias':
        return i
    else:
        if i <= length / 2:
            return i
        else:
            return length - i + 1


def w(AnomalyRange, p):
    MyValue = 0
    MaxValue = 0
    start = AnomalyRange[0]
    AnomalyLength = AnomalyRange[1] - AnomalyRange[0] + 1
    # flat/'front-end bias'/'back-end bias'
    bias = 'flat'
    for i in range(start, start + AnomalyLength):
        bi = b(bias, i, AnomalyLength)
        MaxValue += bi
        if i in p:
            MyValue += bi
    return MyValue / MaxValue


def Cardinality_factor(Anomolyrange, Prange):
    score = 0
    start = Anomolyrange[0]
    end = Anomolyrange[1]
    for i in Prange:
        if start <= i[0] <= end:
            score += 1
        elif i[0] <= start <= i[1]:
            score += 1
        elif i[0] <= end <= i[1]:
            score += 1
        elif start >= i[0] and end <= i[1]:
            score += 1
    if score == 0:
        return 0
    else:
        return 1 / score


def existence_reward(labels, preds):
    '''
    labels: list of ordered pair
    preds predicted data
    '''

    score = 0
    for i in labels:
        if np.sum(np.multiply(preds <= i[1], preds >= i[0])) > 0:
            score += 1
    return score


def range_recall_new(labels, preds, alpha):
    p = np.where(preds == 1)[0]  # positions of predicted label==1
    range_pred = range_convers_new(preds)
    range_label = range_convers_new(labels)

    Nr = len(range_label)  # total # of real anomaly segments

    ExistenceReward = existence_reward(range_label, p)

    OverlapReward = 0
    for i in range_label:
        OverlapReward += w(i, p) * Cardinality_factor(i, range_pred)

    score = alpha * ExistenceReward + (1 - alpha) * OverlapReward
    if Nr != 0:
        return score / Nr, ExistenceReward / Nr, OverlapReward / Nr
    else:
        return 0, 0, 0

# 
def customizable_f1_score(y_test, pred_labels,  alpha=0.2):
    label = y_test
    preds = pred_labels
    Rrecall, ExistenceReward, OverlapReward = range_recall_new(label, preds, alpha)
    Rprecision = range_recall_new(preds, label, 0)[0]

    if Rprecision + Rrecall == 0:
        Rf = 0
    else:
        Rf = 2 * Rrecall * Rprecision / (Rprecision + Rrecall)
    return Rf


def main():
    y_test = np.zeros(100)
    y_test[10:20] = 1
    y_test[50:60] = 1
    pred_labels = np.zeros(100)
    pred_labels[15:19] = 1
    pred_labels[55:62] = 1
    # pred_labels[51:55] = 1
    # true_events = get_events(y_test)
    Rf = customizable_f1_score(y_test, pred_labels)
    print("Rf: {}".format(Rf))


if __name__ == "__main__":
    main()


from sklearn import metrics
import numpy as np
import math
import copy
from itertools import groupby
from operator import itemgetter

METRIC_LIST = ['CCE', 'F1', 'F1-PA', 'Reduced-F1', 'R-based F1', 'eTaPR', 'Aff-F1', 'UAff-F1', 'AUC-ROC', 'VUS-ROC', 'VUS-PR', 'AUC-PR', 'PA%K', 'TaPR', 'PATE']

try:
    from eTaPR_pkg.tapr import print_result, compute
    from eTaPR_pkg.etapr import evaluate_w_ranges
    from eTaPR_pkg.DataManage.File_IO import load_list
    from pate.PATE_metric import PATE
    from affiliation.generics import convert_vector_to_events
    from affiliation.metrics import pr_from_events
except Exception:
    import os
    import sys
    fil_pth = os.path.dirname(os.path.abspath(__file__))
    # proj_dir = os.path.dirname(fil_pth)  # 获取上上级路径
    sys.path.append(fil_pth)  # 将项目根目录添加到系统路径中
    from eTaPR_pkg.tapr import print_result, compute
    from eTaPR_pkg.etapr import evaluate_w_ranges
    from eTaPR_pkg.DataManage.File_IO import load_list
    from pate.PATE_metric import PATE
    from affiliation.generics import convert_vector_to_events
    from affiliation.metrics import pr_from_events



def generate_curve(label, score, slidingWindow, version='opt', thre=250):
    if version =='opt_mem':
        tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = basic_metricor().RangeAUC_volume_opt_mem(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)
    else:
        tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = basic_metricor().RangeAUC_volume_opt(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)

    X = np.array(tpr_3d).reshape(1,-1).ravel()
    X_ap = np.array(tpr_3d)[:,:-1].reshape(1,-1).ravel()
    Y = np.array(fpr_3d).reshape(1,-1).ravel()
    W = np.array(prec_3d).reshape(1,-1).ravel()
    Z = np.repeat(window_3d, len(tpr_3d[0]))
    Z_ap = np.repeat(window_3d, len(tpr_3d[0])-1)

    return Y, Z, X, X_ap, W, Z_ap,avg_auc_3d, avg_ap_3d

def generate_curve_ROC(label, score, slidingWindow, version='opt', thre=250):
    if version =='opt_mem':
        # tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = basic_metricor().RangeAUC_volume_opt_mem(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)
        raise NotImplementedError("opt_mem version is not implemented for ROC curve generation.")
    else:
        tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = basic_metricor().RangeAUC_volume_opt_ROC_only(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)

    X = np.array(tpr_3d).reshape(1,-1).ravel()
    X_ap = np.array(tpr_3d)[:,:-1].reshape(1,-1).ravel()
    Y = np.array(fpr_3d).reshape(1,-1).ravel()
    W = np.array(prec_3d).reshape(1,-1).ravel()
    Z = np.repeat(window_3d, len(tpr_3d[0]))
    Z_ap = np.repeat(window_3d, len(tpr_3d[0])-1)

    return Y, Z, X, X_ap, W, Z_ap,avg_auc_3d, None

def generate_curve_PR(label, score, slidingWindow, version='opt', thre=250):
    if version =='opt_mem':
        # tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = basic_metricor().RangeAUC_volume_opt_mem(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)
        raise NotImplementedError("opt_mem version is not implemented for PR curve generation.")
    else:
        tpr_3d, fpr_3d, prec_3d, window_3d, avg_auc_3d, avg_ap_3d = basic_metricor().RangeAUC_volume_opt_PR_only(labels_original=label, score=score, windowSize=slidingWindow, thre=thre)

    X = np.array(tpr_3d).reshape(1,-1).ravel()
    X_ap = np.array(tpr_3d)[:,:-1].reshape(1,-1).ravel()
    Y = np.array(fpr_3d).reshape(1,-1).ravel()
    W = np.array(prec_3d).reshape(1,-1).ravel()
    Z = np.repeat(window_3d, len(tpr_3d[0]))
    Z_ap = np.repeat(window_3d, len(tpr_3d[0])-1)

    return Y, Z, X, X_ap, W, Z_ap,None, avg_ap_3d

import math,torch

class MyAffBiasCache:
    def __init__(self) -> None:
        self.cache = {}

    def get_label_hash(self, label):
        if isinstance(label, np.ndarray):
            label = label.reshape(-1).tolist()
        elif isinstance(label, torch.Tensor):
            label = label.reshape(-1).detach().cpu().numpy().tolist()

        label = tuple(label)  # Convert to tuple for caching
        label_hash = hash(label)
        return label_hash

    def cache_bias(self, label, bias):
        label_hash = self.get_label_hash(label)
        self.cache[label_hash] = bias

    def get_cached_bias(self, label):
        label_hash = self.get_label_hash(label)
        return self.cache.get(label_hash, None)
    


def convert_vector_to_events(vector):
    """
    Convert a binary vector (indicating 1 for the anomalous instances)
    to a list of events. The events are considered as durations,
    i.e. setting 1 at index i corresponds to an anomalous interval [i, i+1).
    
    :param vector: a list of elements belonging to {0, 1}
    :return: a list of couples, each couple representing the start and stop of
    each event
    """
    positive_indexes = [idx for idx, val in enumerate(vector) if val > 0]
    events = []
    for k, g in groupby(enumerate(positive_indexes), lambda ix : ix[0] - ix[1]):
        cur_cut = list(map(itemgetter(1), g))
        events.append((cur_cut[0], cur_cut[-1]))
    
    # Consistent conversion in case of range anomalies (for indexes):
    # A positive index i is considered as the interval [i, i+1),
    # so the last index should be moved by 1
    events = [(x, y+1) for (x,y) in events]
        
    return (events)


class ConfidenceConsistencyEvaluation:
    # or name "UncertaintyAwareConfidenceConsistencyAnomalyMetric"
    """
    基于不确定度的异常检测评估指标
    
    核心思想：
    1. 量化模型对不同区域的不确定性
    2. 异常区域应该有高置信度高一致性
    3. 正常区域应该有低置信度高一致性
    4. 通过不确定度与真实标签的一致性来评估模型
    """
    
    def __init__(self,  method='bayesian_v2', confidence_level=0.5, n_bootstrap_samples=100, positive_constraint=False,
                 bayesian_scale=10):
        """
        初始化评估器
        
        Args:
            n_bootstrap_samples: 用于估计不确定度的bootstrap样本数
            confidence_level: 置信区间水平, 默认为0.5。
        """
        self.n_bootstrap_samples = n_bootstrap_samples
        self.confidence_level = confidence_level
        self.method = method
        self.positive_constraint = positive_constraint
        self.bayesian_scale = bayesian_scale  # 贝叶斯方法的缩放因子
    
    def estimate_uncertainty(self, model_scores, method='bayesian'):
        """
        估计模型分数的不确定性
        
        Args:
            model_scores: 模型预测分数
            method: 不确定度估计方法 ('bootstrap', 'ensemble', 'bayesian', 'gaussian')
            
        Returns:
            uncertainty_scores: 不确定度分数
        """
        if method == 'bootstrap':
            # TODO, TODEBUG
            return self._bootstrap_uncertainty(model_scores)
        elif method == 'ensemble':
            # TODO, TODEBUG
            return self._ensemble_uncertainty(model_scores)
        elif method == 'bayesian':
            return self._bayesian_uncertainty(model_scores, self.bayesian_scale)
        elif method == 'bayesian_v2':
            return self._bayesian_uncertainty_v2(model_scores, self.bayesian_scale)
        elif method == 'gaussian':
            return self._gaussian_uncertainty(model_scores)
        else:
            raise ValueError(f"Unsupported uncertainty method: {method}")
    
    def _bootstrap_uncertainty(self, model_scores):
        """
        使用bootstrap方法估计不确定度
        """
        n_samples = len(model_scores)
        bootstrap_scores = []
        
        for _ in range(self.n_bootstrap_samples):
            # 随机采样（有放回）
            indices = np.random.choice(n_samples, size=n_samples, replace=True)
            bootstrap_scores.append(model_scores[indices])
        
        bootstrap_scores = np.array(bootstrap_scores)
        
        # 计算每个位置的方差作为不确定度
        uncertainty = np.var(bootstrap_scores, axis=0)
        
        return uncertainty
    
    def _bayesian_uncertainty_v2(self, data, scale=10):
        # 计算样本均值
        mean_x = np.mean(data)
        # 计算样本二阶中心矩
        m2 = np.mean((data - mean_x) ** 2)

        # 估计alpha和beta
        alpha = mean_x * ((mean_x * (1 - mean_x) / m2) - 1) * scale + 1
        beta = (1 - mean_x) * ((mean_x * (1 - mean_x) / m2) - 1) * scale + 1

        # 计算Beta分布的方差
        uncertainty = (alpha * beta) / ((alpha + beta) ** 2 * (alpha + beta + 1))
        return uncertainty
    
    def _ensemble_uncertainty(self, model_scores):
        """
        使用集成方法估计不确定度（模拟多个模型的预测）
        """
        # 模拟多个模型的预测
        ensemble_scores = []
        
        for i in range(self.n_bootstrap_samples):
            # 添加随机噪声模拟不同模型的预测
            noise = np.random.normal(0, 0.1, len(model_scores))
            ensemble_score = model_scores + noise
            ensemble_score = np.clip(ensemble_score, 0, 1)
            ensemble_scores.append(ensemble_score)
        
        ensemble_scores = np.array(ensemble_scores)
        uncertainty = np.var(ensemble_scores, axis=0)
        
        return uncertainty
    
    def _bayesian_uncertainty_v2(self, data, scale=1):
        # 计算样本均值
        mean_x = np.mean(data)
        # 计算样本二阶中心矩
        m2 = np.mean((data - mean_x) ** 2)

        # 估计alpha和beta
        alpha = mean_x * ((mean_x * (1 - mean_x) / (m2+1e-8)) - 1)
        beta = (1 - mean_x) * ((mean_x * (1 - mean_x) / (m2+1e-8)) - 1)

        # 计算Beta分布的方差
        uncertainty = (alpha * beta) / ((alpha + beta) ** 2 * (alpha + beta + 1) + 1e-8)
        return uncertainty
    
    def _bayesian_uncertainty(self, model_scores, scale=10):
        """
        使用贝叶斯方法估计不确定度
        """
        # 使用Beta分布建模分数的不确定性
        # 将分数转换为Beta分布的参数
        alpha = model_scores * scale + 1  # 避免参数为0
        beta = (1 - model_scores) * scale + 1
        
        # 计算Beta分布的方差作为不确定度
        uncertainty = (alpha * beta) / ((alpha + beta) ** 2 * (alpha + beta + 1))
        uncertainty = np.mean(uncertainty)
        return uncertainty
    
    def _gaussian_uncertainty(self, model_scores):
        """
        使用高斯分布估计不确定度
        """
        uncertainty = np.var(model_scores)
        return uncertainty
    
    def metric_score(self, y_true, y_scores):
        """
        计算基于不确定度的异常检测评估分数
        """
        score = self.compute_confidence_consistency_score(y_true, y_scores, self.method)
        return {
            'confidence_consistency_score': score,
        }
        
    def _anom_event_score(self, y_score):
        mu = np.mean(y_score)
        uncertainty = self.estimate_uncertainty(y_score, self.method)
        consistency = np.exp(-uncertainty)
        if self.positive_constraint:
            confidence = max(mu - self.confidence_level,0)
        else:
            confidence = mu - self.confidence_level
        score = confidence * consistency
        return score
    
    def _normal_event_score(self, y_score):
        mu = np.mean(y_score)
        uncertainty = self.estimate_uncertainty(y_score, self.method)
        consistency = np.exp(-uncertainty)
        if self.positive_constraint:
            confidence = max(1 - self.confidence_level - mu,0)
        else:
            confidence = 1 - self.confidence_level - mu
        score = confidence * consistency
        return score
    
    def _event_score(self, anom_uncertainty,normal_uncertainty):
        anom_score = np.mean(anom_uncertainty)
        norm_score = np.mean(normal_uncertainty)
        if self.positive_constraint:
            event_score = 2 * anom_score * norm_score / (anom_score + norm_score + 1e-8)
        else:
            anom_score_ = abs(anom_score)
            norm_score_ = abs(norm_score)
            event_score = 2 * anom_score_ * norm_score_ / (anom_score_ + norm_score_ + 1e-8)
            if anom_score <0 or norm_score < 0:
                event_score *= -1

        return event_score
    
    def _event_score_v2(self, anom_uncertainty,normal_uncertainty, weight=0.5):
        anom_score = np.mean(anom_uncertainty)
        norm_score = np.mean(normal_uncertainty)
        event_score = anom_score * weight + norm_score * (1 - weight)
        return event_score
    
    def _global_score(self, y_true, y_scores):
        anom_list = y_scores[y_true==1]
        norm_list = y_scores[y_true==0]
        glo_anom_score = self._anom_event_score(anom_list)
        glo_norm_score = self._normal_event_score(norm_list)
        if self.positive_constraint:
            glo_score = 2 * glo_anom_score * glo_norm_score / (glo_anom_score + glo_norm_score + 1e-8)
        else:
            glo_anom_score_ = abs(glo_anom_score)
            glo_norm_score_ = abs(glo_norm_score)
            glo_score = 2 * glo_anom_score_ * glo_norm_score_ / (glo_anom_score_ + glo_norm_score_ + 1e-8)
            if glo_anom_score <0 or glo_norm_score < 0:
                glo_score *= -1
        return glo_score

    def _global_score_v2(self, y_true, y_scores, weight=0.5):
        anom_list = y_scores[y_true==1]
        norm_list = y_scores[y_true==0]
        glo_anom_score = self._anom_event_score(anom_list)
        glo_norm_score = self._normal_event_score(norm_list)
        glo_score = glo_anom_score * weight + glo_norm_score * (1 - weight)
        return glo_score
    
    def compute_confidence_consistency_score(self, y_true, y_scores):
        """
        计算基于不确定度的异常检测评估分数
        
        Args:
            y_true: 真实标签
            y_scores: 模型预测分数
            method: 不确定度估计方法
            
        Returns:
            UCE分数, 0-1, 越大保持一致性越好, 结果越可信
        """
        # ytrue只有0或者1
        y_true = y_true.astype(int)
        
        # 获取异常区间 
        anom_events = convert_vector_to_events(y_true)
        normal_events = convert_vector_to_events(1 - y_true)
        
        y_scores = (y_scores - min(y_scores)) / (max(y_scores) - min(y_scores) + 1e-8)
        anom_uncertainty = []
        normal_uncertainty = []
        for st, ed in anom_events:
            anom_uncertainty.append(self._anom_event_score(y_scores[st:ed]))
        for st, ed in normal_events:
            normal_uncertainty.append(self._normal_event_score(y_scores[st:ed]))
        score_event =  self._event_score(anom_uncertainty,normal_uncertainty)
        score_global = self._global_score(y_true,y_scores)
        score = score_event + score_global
        return score
    
    # 更鲁棒的版本
    def compute_confidence_consistency_score_v2(self, y_true, y_scores, weight=0.5):
        """
        计算基于不确定度的异常检测评估分数
        
        Args:
            y_true: 真实标签
            y_scores: 模型预测分数
            method: 不确定度估计方法
            
        Returns:
            UCE分数, 0-1, 越大保持一致性越好, 结果越可信
        """
        # ytrue只有0或者1
        y_true = y_true.astype(int)
        
        # 获取异常区间 
        anom_events = convert_vector_to_events(y_true)
        normal_events = convert_vector_to_events(1 - y_true)
        
        y_scores = (y_scores - min(y_scores)) / (max(y_scores) - min(y_scores) + 1e-8)
        anom_uncertainty = []
        normal_uncertainty = []
        for st, ed in anom_events:
            anom_uncertainty.append(self._anom_event_score(y_scores[st:ed]))
        for st, ed in normal_events:
            normal_uncertainty.append(self._normal_event_score(y_scores[st:ed]))

        score_event =  self._event_score_v2(anom_uncertainty,normal_uncertainty,weight)
        score_global = self._global_score_v2(y_true,y_scores,weight)
        score = score_event + score_global
        return score
    
class basic_metricor():
    def __init__(self, a = 1, probability = True, bias = 'flat', ):
        self.a = a
        self.probability = probability
        self.bias = bias
        self.eps = 1e-15
        self.Unbiased_Aff_prec_bias_cache = MyAffBiasCache()

    def metric_by_name(self, name, labels, score, preds=None, **kwargs):
        """
        根据名称调用对应的指标计算函数。
        
        Args:
            name: 指标名称
            labels: 真实标签
            score: 模型预测分数
            preds: 可选，预测标签
            
        Returns:
            指标计算结果
        """
        if name == 'CCE':
            results = self.metric_CCE(labels, score, **kwargs)
        elif name == 'F1':
            results = self.metric_PointF1(labels, score, preds=preds)
        elif name == 'F1-PA':
            results = self.metric_PointF1PA(labels, score, preds=preds)
        elif name == 'Reduced-F1':
            results = self.metric_Reduced_F1(labels, score, preds=preds)
        elif name == 'R-based F1':
            results = self.metric_RF1(labels, score, preds=preds)
        elif name == 'eTaPR':
            results = self.metric_eTaPR_F1(labels, score, preds=preds, **kwargs)
        elif name == 'PA%K':
            results = self.metric_PA_percentile_K(labels, score, preds=preds, **kwargs)
        elif name == 'TaPR':
            results = self.metric_TaPR_F1(labels, score, preds=preds, **kwargs)
        elif name == 'Aff-F1':
            results = self.metric_Affiliation(labels, score, preds=preds, **kwargs)
        elif name == 'UAff-F1':
            results = self.metric_UN_Affiliation(labels, score, pred=preds, **kwargs)
        elif name == 'AUC-ROC':
            results = self.metric_ROC(labels, score)
        elif name == 'AUC-PR':
            results = self.metric_PR(labels, score)
        elif name == 'VUS-ROC':
            results = self.metric_VUS_ROC(labels, score, **kwargs)
        elif name == 'VUS-PR':
            results = self.metric_VUS_PR(labels, score, **kwargs)
        elif name == 'PATE':
            results = self.metric_PATE(labels, score, **kwargs)
        else:
            raise ValueError(f"Unsupported metric name: {name}")
        
        if isinstance(results, (tuple, list, np.ndarray)):
            first_value = results[0]
        else:
            first_value = results  # 标量
        return first_value

    def get_pred(self, score, quantile=0.95):
        """
        Get the predicted labels based on the score and quantile.
        """
        assert 0 < quantile < 1, "Quantile must be between 0 and 1."
        
        thres = np.percentile(score, quantile*100)
        preds = score > thres
        return preds

    def cal_unbiased_aff_prec_bias(self, label, sample_cnt=3, verbose=1):
        if self.Unbiased_Aff_prec_bias_cache.get_cached_bias(label) is not None:
            if verbose:
                print("[Info]: Unbiased_Aff_prec_bias is already cached. Use the cached value.")
            return self.Unbiased_Aff_prec_bias_cache.get_cached_bias(label)
        else:
            precs_bias = []
            for i in range(sample_cnt):
                rnd_score = np.random.rand(len(label))
                thres = np.percentile(rnd_score, 98)
                preds = rnd_score > thres
                Affiliation_F, Aff_pre, Aff_rec = self.metric_Affiliation(label, rnd_score, preds=preds)
                precs_bias.append(Aff_pre)
            avg_bias = np.mean(precs_bias)
            if verbose:
                print(f"[Info]: Unbiased_Aff_prec_bias is set to {avg_bias:.4f} by sampling {sample_cnt} times.")
                out_str = ", ".join([f"{v:.4}" for i, v in enumerate(precs_bias)])
                print(f"[Info]: Unbiased_Aff_prec_bias sampling results: [{out_str}]")          
                    
            self.Unbiased_Aff_prec_bias_cache.cache_bias(label, avg_bias)
        return avg_bias

    def _metric_PA_percentile_K_k(self, labels, score, preds=None, k = 0.1):
        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            PointF1PA_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)

                adjust_preds = self._adjust_predicts_k(score, labels, pred=preds, k=k)
                PointF1PA = metrics.f1_score(labels, adjust_preds)

                PointF1PA_scores.append(PointF1PA)

            PointF1PA_Threshold = thresholds[np.argmax(PointF1PA_scores)]
            PointF1PA1 = max(PointF1PA_scores)
            preds = score > PointF1PA_Threshold
            adjust_preds = self._adjust_predicts_k(score, labels, pred=preds, k=k)
            adjust_preds = adjust_preds.astype(int)
            precision, recall, f_score, s = metrics.precision_recall_fscore_support(labels, adjust_preds, average='binary')
            assert f_score==PointF1PA1
            Pre = precision
            Rec = recall
        else:
            adjust_preds = self._adjust_predicts_k(score, labels, pred=preds, k=k)
            PointF1PA1 = metrics.f1_score(labels, adjust_preds)

            precision, recall, f_score, s = metrics.precision_recall_fscore_support(labels, adjust_preds, average='binary')
            assert f_score==PointF1PA1
            Pre = precision
            Rec = recall

        F1_Per_K = PointF1PA1
        return F1_Per_K, Pre, Rec
    
    def metric_CCE(self, labels, scores, method='bayesian_v2', confidence_level=0.5, n_samples=30, positive_constraint=False, bayesian_scale=10):
        cce = ConfidenceConsistencyEvaluation(method, confidence_level, n_samples, positive_constraint, bayesian_scale)
        score = cce.compute_confidence_consistency_score_v2(labels, scores)
        return score

    def metric_PA_percentile_K(self, labels, score, preds=None, num_K=100):
        if preds is None:
            print("[Warning]: If the preds is None, the calculation will be time-consuming!!! We will calculate the threshold for each K percentile. Please set the preds manually if you want to speed up the calculation.")
        thresholds = np.linspace(0, 1, num_K)
        F1_Per_Ks = []
        Prec_Per_Ks = []
        Rec_Per_Ks = []
        for k in thresholds:
            F1_Per_K, Pre, Rec = self._metric_PA_percentile_K_k(labels, score, preds=preds, k=k)
            F1_Per_Ks.append(F1_Per_K)
            Prec_Per_Ks.append(Pre)
            Rec_Per_Ks.append(Rec)
        F1_Per_K = np.mean(F1_Per_Ks)
        Prec_Per_K = np.mean(Prec_Per_Ks)
        Rec_Per_K = np.mean(Rec_Per_Ks)
        return F1_Per_K, Prec_Per_K, Rec_Per_K
        

    def metric_PATE(self, labels, score, e_buffer=100, d_buffer=100, n_jobs=100, num_desired_thresholds=250):
        result = PATE(labels, score, e_buffer=e_buffer, d_buffer=d_buffer, n_jobs=n_jobs, num_desired_thresholds= num_desired_thresholds)
        return result

    def _metric_eTaPR_F1_pred(self, labels, preds, theta_p=0.5, theta_r=0.1, delta=0.0):
        anomalies = load_list(labels)
        predictions = load_list(preds)
        result = evaluate_w_ranges(anomalies, predictions, theta_p, theta_r, delta)
        eTaP = result['eTaP']
        eTaR = result['eTaR']
        eTaF1 = result['f1']
        return eTaP, eTaR, eTaF1

    def _metric_TaPR_F1_pred(self, labels, preds, alpha=0.6, theta=0.8, delta=600):
        anomalies = load_list(labels)
        predictions = load_list(preds)
        result = compute(anomalies, predictions, alpha, theta, delta)
        TaP = result['TaP']
        TaR = result['TaR']
        TaF1 = result['f1']
        return TaP, TaR, TaF1
    
    def metric_eTaPR_F1(self, labels, score=None, preds=None, theta_p=0.5, theta_r=0.1, delta=0.0):
        if score is None and preds is None:
            raise ValueError("Either score or preds must be provided.")
        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            eTaPR_F1_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)
                eTaP, eTaR, eTaF1 = self._metric_eTaPR_F1_pred(labels, preds, theta_p=theta_p, theta_r=theta_r, delta=delta)
                eTaPR_F1_scores.append(eTaF1)

            eTaPR_F1_Threshold = thresholds[np.argmax(eTaPR_F1_scores)]
            eTaPR_F1 = max(eTaPR_F1_scores)
            preds = score > eTaPR_F1_Threshold
            eTaP, eTaR, eTaF1 = self._metric_eTaPR_F1_pred(labels, preds, theta_p=theta_p, theta_r=theta_r, delta=delta)

        else:
            eTaP, eTaR, eTaF1 = self._metric_eTaPR_F1_pred(labels, preds, theta_p=theta_p, theta_r=theta_r, delta=delta)

        return  eTaF1, eTaP, eTaR

    def metric_TaPR_F1(self, labels, score=None, preds=None, alpha=0.6, theta=0.8, delta=600):
        if score is None and preds is None:
            raise ValueError("Either score or preds must be provided.")
        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            TaPR_F1_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)
                TaP, TaR, TaF1 = self._metric_TaPR_F1_pred(labels, preds, alpha=alpha, theta=theta, delta=delta)
                TaPR_F1_scores.append(TaF1)

            TaPR_F1_Threshold = thresholds[np.argmax(TaPR_F1_scores)]
            TaPR_F1 = max(TaPR_F1_scores)
            preds = score > TaPR_F1_Threshold
            TaP, TaR, TaF1 = self._metric_TaPR_F1_pred(labels, preds, alpha=alpha, theta=theta, delta=delta)
        else:
            TaP, TaR, TaF1 = self._metric_TaPR_F1_pred(labels, preds, alpha=alpha, theta=theta, delta=delta)
        return TaF1, TaP, TaR

    def _adjust_length(self, labels, preds, base=3):
        anom_ranges = self.range_convers_new(labels)
        new_label = []
        new_preds = []
        for start, end in anom_ranges:
            new_len = end - start
            new_len = math.floor(math.log(new_len+base, base))
            new_label.extend([1] * new_len)
            new_preds.extend([1] * new_len if preds[start] else [0] * new_len)
        norm_tmp = labels.copy()
        norm_tmp[norm_tmp == 0] = -1
        norm_tmp[norm_tmp == 1] = 0
        norm_tmp[norm_tmp == -1] = 1
        norm_ranges = self.range_convers_new(norm_tmp)
        norm_ranges = convert_vector_to_events(norm_tmp)
        for start, end in norm_ranges:
            lenz = end - start
            new_label.extend([0] * lenz)
            new_preds.extend(preds[start:end].tolist())

        return np.array(new_label), np.array(new_preds)

    def metric_Reduced_F1(self, labels, score, preds=None):
        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            PointF1PA_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)
                adjust_preds = self._adjust_predicts(score, labels, pred=preds)
                PointF1PA = metrics.f1_score(labels, adjust_preds)
                PointF1PA_scores.append(PointF1PA)

            PointF1PA_Threshold = thresholds[np.argmax(PointF1PA_scores)]
            # PointF1PA1 = max(PointF1PA_scores)
            preds = score > PointF1PA_Threshold
            adjust_preds = self._adjust_predicts(score, labels, pred=preds)
            adjust_preds = adjust_preds.astype(int)
            score, adjust_preds = self._adjust_length(labels, adjust_preds)
            precision, recall, f_score, s = metrics.precision_recall_fscore_support(labels, adjust_preds, average='binary')
            Reduced_F1 = f_score
            Pre = precision
            Rec = recall

        else:
            adjust_preds = self._adjust_predicts(score, labels, pred=preds)
            labels, adjust_preds = self._adjust_length(labels, adjust_preds)
            precision, recall, f_score, s = metrics.precision_recall_fscore_support(labels, adjust_preds, average='binary')
            Reduced_F1 = f_score
            Pre = precision
            Rec = recall

        return Reduced_F1, Pre, Rec

        
    def metric_UN_Affiliation(self, labels, score, pred=None):
        Affiliation_F, Aff_pre, Aff_rec = self.metric_Affiliation(labels, score, preds=pred)
        bias_aff_pre = self.Unbiased_Aff_prec_bias_cache.get_cached_bias(labels)
        if bias_aff_pre is None:
            print("[Info]: Unbiased_Aff_prec_bias is not set. Calculate by sample bias. The time-cost will be higher than setting it manually. You can set it by calling `cal_unbiased_aff_prec_bias` function.")
            bias_aff_pre = self.cal_unbiased_aff_prec_bias(labels)
        UAff_Pre = (Aff_pre - bias_aff_pre)/(1-bias_aff_pre)
        UAff_F1 = 2*abs(UAff_Pre)*Aff_rec/(abs(UAff_Pre)+Aff_rec)
        if UAff_Pre < 0:
            UAff_F1 *= -1

        NAff_Pre = (Aff_pre - 0.5)/(1-0.5)
        NAff_F1 = 2*abs(NAff_Pre)*Aff_rec/(abs(NAff_Pre)+Aff_rec)
        if NAff_Pre < 0:
            NAff_F1 *= -1

        return UAff_F1, NAff_F1

    def metric_U_Affiliation_f1_pre_rec(self, labels, score, pred=None):
        Affiliation_F, Aff_pre, Aff_rec = self.metric_Affiliation(labels, score, preds=pred)
        bias_aff_pre = self.Unbiased_Aff_prec_bias_cache.get_cached_bias(labels)
        if bias_aff_pre is None:
            print("[Info]: Unbiased_Aff_prec_bias is not set. Calculate by sample bias. The time-cost will be higher than setting it manually. You can set it by calling `cal_unbiased_aff_prec_bias` function.")
            bias_aff_pre = self.cal_unbiased_aff_prec_bias(labels)
        UAff_Pre = (Aff_pre - bias_aff_pre)/(1-bias_aff_pre)
        UAff_F1 = 2*abs(UAff_Pre)*Aff_rec/(abs(UAff_Pre)+Aff_rec)
        if UAff_Pre < 0:
            UAff_F1 *= -1

        return UAff_F1, UAff_Pre, Aff_rec
    
    def metric_N_Affiliation_f1_pre_rec(self, labels, score, pred=None):
        Affiliation_F, Aff_pre, Aff_rec = self.metric_Affiliation(labels, score, preds=pred)
        NAff_Pre = (Aff_pre - 0.5)/(1-0.5)
        NAff_F1 = 2*abs(NAff_Pre)*Aff_rec/(abs(NAff_Pre)+Aff_rec)
        if NAff_Pre < 0:
            NAff_F1 *= -1

        return NAff_F1, NAff_Pre, Aff_rec

    def detect_model(self, model, label, contamination = 0.1, window = 100, is_A = False, is_threshold = True):
        if is_threshold:
            score = self.scale_threshold(model.decision_scores_, model._mu, model._sigma)
        else:
            score = self.scale_contamination(model.decision_scores_, contamination = contamination)
        if is_A is False:
            scoreX = np.zeros(len(score)+window)
            scoreX[math.ceil(window/2): len(score)+window - math.floor(window/2)] = score
        else:
            scoreX = score

        self.score_=scoreX
        L = self.metric(label, scoreX)
        return L
    
    def metric_VUS_ROC_PR(self, labels, scores, slidingWindow=100, version='opt', thre=250):
        _, _, _, _, _, _,VUS_ROC, VUS_PR = generate_curve(labels.astype(int), scores, 2*slidingWindow, version, thre)
        return VUS_ROC, VUS_PR
    
    def metric_VUS_ROC(self, labels, scores, slidingWindow=100, version='opt', thre=250):
        _, _, _, _, _, _,VUS_ROC, VUS_PR = generate_curve_ROC(labels.astype(int), scores, 2*slidingWindow, version, thre)
        return VUS_ROC
    
    def metric_VUS_PR(self, labels, scores, slidingWindow=100, version='opt', thre=250):
        _, _, _, _, _, _,VUS_ROC, VUS_PR = generate_curve_PR(labels.astype(int), scores, 2*slidingWindow, version, thre)
        return VUS_PR

    def w(self, AnomalyRange, p):
        MyValue = 0
        MaxValue = 0
        start = AnomalyRange[0]
        AnomalyLength = AnomalyRange[1] - AnomalyRange[0] + 1
        for i in range(start, start +AnomalyLength):
            bi = self.b(i, AnomalyLength)
            MaxValue +=  bi
            if i in p:
                MyValue += bi
        return MyValue/MaxValue

    def Cardinality_factor(self, Anomolyrange, Prange):
        score = 0
        start = Anomolyrange[0]
        end = Anomolyrange[1]
        for i in Prange:
            if i[0] >= start and i[0] <= end:
                score +=1
            elif start >= i[0] and start <= i[1]:
                score += 1
            elif end >= i[0] and end <= i[1]:
                score += 1
            elif start >= i[0] and end <= i[1]:
                score += 1
        if score == 0:
            return 0
        else:
            return 1/score

    def b(self, i, length):
        bias = self.bias
        if bias == 'flat':
            return 1
        elif bias == 'front-end bias':
            return length - i + 1
        elif bias == 'back-end bias':
            return i
        else:
            if i <= length/2:
                return i
            else:
                return length - i + 1

    def scale_threshold(self, score, score_mu, score_sigma):
        return (score >= (score_mu + 3*score_sigma)).astype(int)

    def _adjust_predicts_k(self, score, label, threshold=None, pred=None, k=0.):
        if len(score) != len(label):
            raise ValueError("score and label must have the same length")
        score = np.asarray(score)
        label = np.asarray(label)
        if pred is None:
            predict = score > threshold
        else:
            predict = copy.deepcopy(pred)
        anom_ranges = self.range_convers_new(label)
        for st, ed in anom_ranges:
            pred_corr = sum(predict[st:ed])
            proportion = pred_corr / (ed - st)
            if proportion >= k:
                predict[st:ed] = 1
        return predict
        
        

    def _adjust_predicts(self, score, label, threshold=None, pred=None, calc_latency=False):
        """
        Calculate adjusted predict labels using given `score`, `threshold` (or given `pred`) and `label`.

        Args:
            score (np.ndarray): The anomaly score
            label (np.ndarray): The ground-truth label
            threshold (float): The threshold of anomaly score.
                A point is labeled as "anomaly" if its score is higher than the threshold.
            pred (np.ndarray or None): if not None, adjust `pred` and ignore `score` and `threshold`,
            calc_latency (bool):

        Returns:
            np.ndarray: predict labels
        """
        if len(score) != len(label):
            raise ValueError("score and label must have the same length")
        score = np.asarray(score)
        label = np.asarray(label)
        latency = 0
        if pred is None:
            predict = score > threshold
        else:
            predict = copy.deepcopy(pred)
        actual = label > 0.1
        anomaly_state = False
        anomaly_count = 0
        for i in range(len(score)):
            if actual[i] and predict[i] and not anomaly_state:
                    anomaly_state = True
                    anomaly_count += 1
                    for j in range(i, 0, -1):
                        if not actual[j]:
                            break
                        else:
                            if not predict[j]:
                                predict[j] = True
                                latency += 1
            elif not actual[i]:
                anomaly_state = False
            if anomaly_state:
                predict[i] = True
        if calc_latency:
            return predict, latency / (anomaly_count + 1e-4)
        else:
            return predict

    def metric_new(self, label, score, preds, plot_ROC=False, alpha=0.2):
        '''input:
               Real labels and anomaly score in prediction

           output:
               AUC,
               Precision,
               Recall,
               F-score,
               Range-precision,
               Range-recall,
               Range-Fscore,
               Precison@k,

            k is chosen to be # of outliers in real labels
        '''
        if np.sum(label) == 0:
            print('All labels are 0. Label must have groud truth value for calculating AUC score.')
            return None

        if np.isnan(score).any() or score is None:
            print('Score must not be none.')
            return None

        #area under curve
        auc = metrics.roc_auc_score(label, score)
        # plor ROC curve
        if plot_ROC:
            fpr, tpr, thresholds  = metrics.roc_curve(label, score)
            # display = metrics.RocCurveDisplay(fpr=fpr, tpr=tpr, roc_auc=auc)
            # display.plot()

        #precision, recall, F
        if preds is None:
            preds = score > (np.mean(score)+3*np.std(score))
        Precision, Recall, F, Support = metrics.precision_recall_fscore_support(label, preds, zero_division=0)
        precision = Precision[1]
        recall = Recall[1]
        f = F[1]

        #point-adjust
        adjust_preds = self._adjust_predicts(score, label, pred=preds)
        PointF1PA = metrics.f1_score(label, adjust_preds)

        #range anomaly
        Rrecall, ExistenceReward, OverlapReward = self.range_recall_new(label, preds, alpha)
        Rprecision = self.range_recall_new(preds, label, 0)[0]

        if Rprecision + Rrecall==0:
            Rf=0
        else:
            Rf = 2 * Rrecall * Rprecision / (Rprecision + Rrecall)

        # top-k
        k = int(np.sum(label))
        threshold = np.percentile(score, 100 * (1-k/len(label)))

        # precision_at_k = metrics.top_k_accuracy_score(label, score, k)
        p_at_k = np.where(preds > threshold)[0]
        TP_at_k = sum(label[p_at_k])
        precision_at_k = TP_at_k/k

        L = [auc, precision, recall, f, PointF1PA, Rrecall, ExistenceReward, OverlapReward, Rprecision, Rf, precision_at_k]
        if plot_ROC:
            return L, fpr, tpr
        return L

    def metric_ROC(self, label, score):
        return metrics.roc_auc_score(label, score)

    def metric_PR(self, label, score):
        return metrics.average_precision_score(label, score)
    
    def metric_PointF1(self, label, score, preds=None):
        if preds is None:
            precision, recall, thresholds = metrics.precision_recall_curve(label, score)
            f1_scores = 2 * (precision * recall) / (precision + recall + 0.00001)
            F1 = np.max(f1_scores)
            threshold = thresholds[np.argmax(f1_scores)]
        else:
            Precision, Recall, F, Support = metrics.precision_recall_fscore_support(label, preds, zero_division=0)
            F1 = F[1]
        return F1


    def metric_Affiliation(self, label, score, preds=None):
        

        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            Affiliation_scores = []
            Affiliation_pre = []
            Affiliation_rec = []
            print("[Warning]: If the preds is None, the calculation will be time-consuming!!! We will calculate the threshold for each percentile. Please set the preds manually if you want to speed up the calculation.")

            for threshold in thresholds:
                preds = (score > threshold).astype(int)

                events_pred = self.range_convers_new(preds)
                events_gt = self.range_convers_new(label)
                Trange = (0, len(preds))
                affiliation_metrics = pr_from_events(events_pred, events_gt, Trange)
                Affiliation_Precision = affiliation_metrics['Affiliation_Precision']
                Affiliation_Recall = affiliation_metrics['Affiliation_Recall']
                Affiliation_F = 2*Affiliation_Precision*Affiliation_Recall / (Affiliation_Precision+Affiliation_Recall+self.eps)

                if Affiliation_F == np.nan or Affiliation_F == np.inf or Affiliation_Precision == 0 or Affiliation_Recall == np.nan\
                    or Affiliation_F ==0 or math.isnan(Affiliation_F) or math.isinf(Affiliation_F) or math.isnan(Affiliation_Precision) or math.isnan(Affiliation_Recall):
                    pass
                else:
                    Affiliation_scores.append(Affiliation_F)
                    Affiliation_pre.append(Affiliation_Precision)
                    Affiliation_rec.append(Affiliation_Recall)

            Affiliation_F1_Threshold = thresholds[np.argmax(Affiliation_scores)]
            Affiliation_F1 = Affiliation_scores[np.argmax(Affiliation_scores)]
            Affiliation_Precision = Affiliation_pre[np.argmax(Affiliation_scores)]
            Affiliation_Recall = Affiliation_rec[np.argmax(Affiliation_scores)]


        else:
            events_pred = self.range_convers_new(preds)
            events_gt = self.range_convers_new(label)
            Trange = (0, len(preds))
            affiliation_metrics = pr_from_events(events_pred, events_gt, Trange)
            Affiliation_Precision = affiliation_metrics['Affiliation_Precision']
            Affiliation_Recall = affiliation_metrics['Affiliation_Recall']
            Affiliation_F1 = 2*Affiliation_Precision*Affiliation_Recall / (Affiliation_Precision+Affiliation_Recall+self.eps)

        return Affiliation_F1, Affiliation_Precision, Affiliation_Recall


    # range-based F1
    def metric_RF1(self, label, score, preds=None):

        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            Rf1_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)

                Rrecall, ExistenceReward, OverlapReward = self.range_recall_new(label, preds, alpha=0.2)
                Rprecision = self.range_recall_new(preds, label, 0)[0]
                if Rprecision + Rrecall==0:
                    Rf=0
                else:
                    Rf = 2 * Rrecall * Rprecision / (Rprecision + Rrecall)

                Rf1_scores.append(Rf)

            RF1_Threshold = thresholds[np.argmax(Rf1_scores)]
            RF1 = max(Rf1_scores)
        else:
            Rrecall, ExistenceReward, OverlapReward = self.range_recall_new(label, preds, alpha=0.2)
            Rprecision = self.range_recall_new(preds, label, 0)[0]
            if Rprecision + Rrecall==0:
                RF1=0
            else:
                RF1 = 2 * Rrecall * Rprecision / (Rprecision + Rrecall)
        return RF1

    # F1-PA
    def metric_PointF1PA(self, label, score, preds=None):
        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            PointF1PA_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)

                adjust_preds = self._adjust_predicts(score, label, pred=preds)
                PointF1PA = metrics.f1_score(label, adjust_preds)

                PointF1PA_scores.append(PointF1PA)

            PointF1PA_Threshold = thresholds[np.argmax(PointF1PA_scores)]
            PointF1PA1 = max(PointF1PA_scores)
            preds = score > PointF1PA_Threshold
            adjust_preds = self._adjust_predicts(score, label, pred=preds)
            adjust_preds = adjust_preds.astype(int)
            # Acc = metrics.accuracy_score(label, adjust_preds)
            precision, recall, f_score, s = metrics.precision_recall_fscore_support(label, adjust_preds, average='binary')
            assert f_score==PointF1PA1
            Pre = precision
            Rec = recall


        else:
            adjust_preds = self._adjust_predicts(score, label, pred=preds)
            PointF1PA1 = metrics.f1_score(label, adjust_preds)

            # Acc = metrics.accuracy_score(label, adjust_preds)
            precision, recall, f_score, s = metrics.precision_recall_fscore_support(label, adjust_preds, average='binary')
            assert f_score==PointF1PA1
            Pre = precision
            Rec = recall

        return PointF1PA1, Pre, Rec

    def _get_events(self, y_test, outlier=1, normal=0):
        events = dict()
        label_prev = normal
        event = 0  # corresponds to no event
        event_start = 0
        for tim, label in enumerate(y_test):
            if label == outlier:
                if label_prev == normal:
                    event += 1
                    event_start = tim
            else:
                if label_prev == outlier:
                    event_end = tim - 1
                    events[event] = (event_start, event_end)
            label_prev = label

        if label_prev == outlier:
            event_end = tim - 1
            events[event] = (event_start, event_end)
        return events

    def metric_EventF1PA(self, label, score, preds=None):
        from sklearn.metrics import precision_score
        true_events = self._get_events(label)

        if preds is None:
            thresholds = np.linspace(score.min(), score.max(), 100)
            EventF1PA_scores = []

            for threshold in thresholds:
                preds = (score > threshold).astype(int)

                tp = np.sum([preds[start:end + 1].any() for start, end in true_events.values()])
                fn = len(true_events) - tp
                rec_e = tp/(tp + fn)
                prec_t = precision_score(label, preds)
                EventF1PA = 2 * rec_e * prec_t / (rec_e + prec_t + self.eps)

                EventF1PA_scores.append(EventF1PA)

            EventF1PA_Threshold = thresholds[np.argmax(EventF1PA_scores)]
            EventF1PA1 = max(EventF1PA_scores)

        else:

            tp = np.sum([preds[start:end + 1].any() for start, end in true_events.values()])
            fn = len(true_events) - tp
            rec_e = tp/(tp + fn)
            prec_t = precision_score(label, preds)
            EventF1PA1 = 2 * rec_e * prec_t / (rec_e + prec_t + self.eps)

        return EventF1PA1

    def range_recall_new(self, labels, preds, alpha):
        p = np.where(preds == 1)[0]    # positions of predicted label==1
        range_pred = self.range_convers_new(preds)
        range_label = self.range_convers_new(labels)

        Nr = len(range_label)    # total # of real anomaly segments

        ExistenceReward = self.existence_reward(range_label, preds)

        OverlapReward = 0
        for i in range_label:
            OverlapReward += self.w(i, p) * self.Cardinality_factor(i, range_pred)


        score = alpha * ExistenceReward + (1-alpha) * OverlapReward
        if Nr != 0:
            return score/Nr, ExistenceReward/Nr, OverlapReward/Nr
        else:
            return 0,0,0

    def range_convers_new(self, label):
        '''
        input: arrays of binary values
        output: list of ordered pair [[a0,b0], [a1,b1]... ] of the inputs
        '''
        # anomaly_starts = np.where(np.diff(label) == 1)[0] + 1
        # anomaly_ends, = np.where(np.diff(label) == -1)
        # if len(anomaly_ends):
        #     if not len(anomaly_starts) or anomaly_ends[0] < anomaly_starts[0]:
        #         # we started with an anomaly, so the start of the first anomaly is the start of the labels
        #         anomaly_starts = np.concatenate([[0], anomaly_starts])
        # if len(anomaly_starts):
        #     if not len(anomaly_ends) or anomaly_ends[-1] < anomaly_starts[-1]:
        #         # we ended on an anomaly, so the end of the last anomaly is the end of the labels
        #         anomaly_ends = np.concatenate([anomaly_ends, [len(label) - 1]])
        # return list(zip(anomaly_starts, anomaly_ends))

        # EmorZz1G 2024-6-25 fix-bug
        return convert_vector_to_events(label)

    def existence_reward(self, labels, preds):
        '''
        labels: list of ordered pair
        preds predicted data
        '''

        score = 0
        for i in labels:
            if preds[i[0]:i[1]+1].any():
                score += 1
        return score

    def num_nonzero_segments(self, x):
        count=0
        if x[0]>0:
            count+=1
        for i in range(1, len(x)):
            if x[i]>0 and x[i-1]==0:
                count+=1
        return count

    def extend_postive_range(self, x, window=5):
        label = x.copy().astype(float)
        L = self.range_convers_new(label)   # index of non-zero segments
        length = len(label)
        for k in range(len(L)):
            s = L[k][0]
            e = L[k][1]


            x1 = np.arange(e,min(e+window//2,length))
            label[x1] += np.sqrt(1 - (x1-e)/(window))

            x2 = np.arange(max(s-window//2,0),s)
            label[x2] += np.sqrt(1 - (s-x2)/(window))

        label = np.minimum(np.ones(length), label)
        return label

    def extend_postive_range_individual(self, x, percentage=0.2):
        label = x.copy().astype(float)
        L = self.range_convers_new(label)   # index of non-zero segments
        length = len(label)
        for k in range(len(L)):
            s = L[k][0]
            e = L[k][1]

            l0 = int((e-s+1)*percentage)

            x1 = np.arange(e,min(e+l0,length))
            label[x1] += np.sqrt(1 - (x1-e)/(2*l0))

            x2 = np.arange(max(s-l0,0),s)
            label[x2] += np.sqrt(1 - (s-x2)/(2*l0))

        label = np.minimum(np.ones(length), label)
        return label

    def TPR_FPR_RangeAUC(self, labels, pred, P, L):
        indices = np.where(labels == 1)[0]
        product = labels * pred
        TP = np.sum(product)
        newlabels = product.copy()
        newlabels[indices] = 1

        # recall = min(TP/P,1)
        P_new = (P + np.sum(newlabels)) / 2  # so TPR is neither large nor small
        # P_new = np.sum(labels)
        recall = min(TP / P_new, 1)
        # recall = TP/np.sum(labels)
        # print('recall '+str(recall))

        existence = 0
        for seg in L:
            if np.sum(product[seg[0]:(seg[1] + 1)]) > 0:  # if newlabels>0, that segment must contained
                existence += 1

        existence_ratio = existence / len(L)
        # print(existence_ratio)

        # TPR_RangeAUC = np.sqrt(recall*existence_ratio)
        # print(existence_ratio)
        TPR_RangeAUC = recall * existence_ratio

        FP = np.sum(pred) - TP
        # TN = np.sum((1-pred) * (1-labels))

        # FPR_RangeAUC = FP/(FP+TN)
        N_new = len(labels) - P_new
        FPR_RangeAUC = FP / N_new

        Precision_RangeAUC = TP / np.sum(pred)

        return TPR_RangeAUC, FPR_RangeAUC, Precision_RangeAUC

    def RangeAUC(self, labels, score, window=0, percentage=0, plot_ROC=False, AUC_type='window'):
        # AUC_type='window'/'percentage'
        score_sorted = -np.sort(-score)

        P = np.sum(labels)
        # print(np.sum(labels))
        if AUC_type == 'window':
            labels = self.extend_postive_range(labels, window=window)
        else:
            labels = self.extend_postive_range_individual(labels, percentage=percentage)

        # print(np.sum(labels))
        L = self.range_convers_new(labels)
        TPR_list = [0]
        FPR_list = [0]
        Precision_list = [1]

        for i in np.linspace(0, len(score) - 1, 250).astype(int):
            threshold = score_sorted[i]
            # print('thre='+str(threshold))
            pred = score >= threshold
            TPR, FPR, Precision = self.TPR_FPR_RangeAUC(labels, pred, P, L)

            TPR_list.append(TPR)
            FPR_list.append(FPR)
            Precision_list.append(Precision)

        TPR_list.append(1)
        FPR_list.append(1)  # otherwise, range-AUC will stop earlier than (1,1)

        tpr = np.array(TPR_list)
        fpr = np.array(FPR_list)
        prec = np.array(Precision_list)

        width = fpr[1:] - fpr[:-1]
        height = (tpr[1:] + tpr[:-1]) / 2
        AUC_range = np.sum(width * height)

        width_PR = tpr[1:-1] - tpr[:-2]
        height_PR = prec[1:]
        AP_range = np.sum(width_PR * height_PR)

        if plot_ROC:
            return AUC_range, AP_range, fpr, tpr, prec

        return AUC_range


    def new_sequence(self, label, sequence_original, window):
        a = max(sequence_original[0][0] - window // 2, 0)
        sequence_new = []
        for i in range(len(sequence_original) - 1):
            if sequence_original[i][1] + window // 2 < sequence_original[i + 1][0] - window // 2:
                sequence_new.append((a, sequence_original[i][1] + window // 2))
                a = sequence_original[i + 1][0] - window // 2
        sequence_new.append((a, min(sequence_original[len(sequence_original) - 1][1] + window // 2, len(label) - 1)))
        return sequence_new

    def sequencing(self, x, L, window=5):
        label = x.copy().astype(float)
        length = len(label)

        for k in range(len(L)):
            s = L[k][0]
            e = L[k][1]

            x1 = np.arange(e + 1, min(e + window // 2 + 1, length))
            label[x1] += np.sqrt(1 - (x1 - e) / (window))

            x2 = np.arange(max(s - window // 2, 0), s)
            label[x2] += np.sqrt(1 - (s - x2) / (window))

        label = np.minimum(np.ones(length), label)
        return label

    # TPR_FPR_window
    def RangeAUC_volume_opt(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d = np.zeros((windowSize + 1, thre + 2))
        fpr_3d = np.zeros((windowSize + 1, thre + 2))
        prec_3d = np.zeros((windowSize + 1, thre + 1))

        auc_3d = np.zeros(windowSize + 1)
        ap_3d = np.zeros(windowSize + 1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)

        for k, i in enumerate(np.linspace(0, len(score) - 1, thre).astype(int)):
            threshold = score_sorted[i]
            pred = score >= threshold
            N_pred[k] = np.sum(pred)

        for window in window_3d:

            labels_extended = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels_extended, seq, window)

            TF_list = np.zeros((thre + 2, 2))
            Precision_list = np.ones(thre + 1)
            j = 0

            for i in np.linspace(0, len(score) - 1, thre).astype(int):
                threshold = score_sorted[i]
                pred = score >= threshold
                labels = labels_extended.copy()
                existence = 0

                for seg in L:
                    labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * pred[seg[0]:seg[1] + 1]
                    if (pred[seg[0]:(seg[1] + 1)] > 0).any():
                        existence += 1
                for seg in seq:
                    labels[seg[0]:seg[1] + 1] = 1

                TP = 0
                N_labels = 0
                for seg in l:
                    TP += np.dot(labels[seg[0]:seg[1] + 1], pred[seg[0]:seg[1] + 1])
                    N_labels += np.sum(labels[seg[0]:seg[1] + 1])

                TP += tp[j]
                FP = N_pred[j] - TP

                existence_ratio = existence / len(L)

                P_new = (P + N_labels) / 2
                recall = min(TP / P_new, 1)

                TPR = recall * existence_ratio
                N_new = len(labels) - P_new
                FPR = FP / N_new

                Precision = TP / N_pred[j]

                j += 1
                TF_list[j] = [TPR, FPR]
                Precision_list[j] = Precision

            TF_list[j + 1] = [1, 1]  # otherwise, range-AUC will stop earlier than (1,1)

            tpr_3d[window] = TF_list[:, 0]
            fpr_3d[window] = TF_list[:, 1]
            prec_3d[window] = Precision_list

            width = TF_list[1:, 1] - TF_list[:-1, 1]
            height = (TF_list[1:, 0] + TF_list[:-1, 0]) / 2
            AUC_range = np.dot(width, height)
            auc_3d[window] = (AUC_range)

            width_PR = TF_list[1:-1, 0] - TF_list[:-2, 0]
            height_PR = Precision_list[1:]

            AP_range = np.dot(width_PR, height_PR)
            ap_3d[window] = AP_range

        return tpr_3d, fpr_3d, prec_3d, window_3d, sum(auc_3d) / len(window_3d), sum(ap_3d) / len(window_3d)
    
    # TPR_FPR_window
    def RangeAUC_volume_opt_PR_only(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d = np.zeros((windowSize + 1, thre + 2))
        fpr_3d = np.zeros((windowSize + 1, thre + 2))
        prec_3d = np.zeros((windowSize + 1, thre + 1))

        # auc_3d = np.zeros(windowSize + 1)
        ap_3d = np.zeros(windowSize + 1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)

        for k, i in enumerate(np.linspace(0, len(score) - 1, thre).astype(int)):
            threshold = score_sorted[i]
            pred = score >= threshold
            N_pred[k] = np.sum(pred)

        for window in window_3d:

            labels_extended = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels_extended, seq, window)

            TF_list = np.zeros((thre + 2, 2))
            Precision_list = np.ones(thre + 1)
            j = 0

            for i in np.linspace(0, len(score) - 1, thre).astype(int):
                threshold = score_sorted[i]
                pred = score >= threshold
                labels = labels_extended.copy()
                existence = 0

                for seg in L:
                    labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * pred[seg[0]:seg[1] + 1]
                    if (pred[seg[0]:(seg[1] + 1)] > 0).any():
                        existence += 1
                for seg in seq:
                    labels[seg[0]:seg[1] + 1] = 1

                TP = 0
                N_labels = 0
                for seg in l:
                    TP += np.dot(labels[seg[0]:seg[1] + 1], pred[seg[0]:seg[1] + 1])
                    N_labels += np.sum(labels[seg[0]:seg[1] + 1])

                TP += tp[j]
                FP = N_pred[j] - TP

                existence_ratio = existence / len(L)

                P_new = (P + N_labels) / 2
                recall = min(TP / P_new, 1)

                TPR = recall * existence_ratio
                N_new = len(labels) - P_new
                FPR = FP / N_new

                Precision = TP / N_pred[j]

                j += 1
                TF_list[j] = [TPR, FPR]
                Precision_list[j] = Precision

            TF_list[j + 1] = [1, 1]  # otherwise, range-AUC will stop earlier than (1,1)

            tpr_3d[window] = TF_list[:, 0]
            fpr_3d[window] = TF_list[:, 1]
            prec_3d[window] = Precision_list

            # width = TF_list[1:, 1] - TF_list[:-1, 1]
            # height = (TF_list[1:, 0] + TF_list[:-1, 0]) / 2
            # AUC_range = np.dot(width, height)
            # auc_3d[window] = (AUC_range)

            width_PR = TF_list[1:-1, 0] - TF_list[:-2, 0]
            height_PR = Precision_list[1:]

            AP_range = np.dot(width_PR, height_PR)
            ap_3d[window] = AP_range

        return tpr_3d, fpr_3d, prec_3d, window_3d, None, sum(ap_3d) / len(window_3d)

    # TPR_FPR_window
    def RangeAUC_volume_opt_ROC_only(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d = np.zeros((windowSize + 1, thre + 2))
        fpr_3d = np.zeros((windowSize + 1, thre + 2))
        prec_3d = np.zeros((windowSize + 1, thre + 1))

        auc_3d = np.zeros(windowSize + 1)
        # ap_3d = np.zeros(windowSize + 1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)

        for k, i in enumerate(np.linspace(0, len(score) - 1, thre).astype(int)):
            threshold = score_sorted[i]
            pred = score >= threshold
            N_pred[k] = np.sum(pred)

        for window in window_3d:

            labels_extended = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels_extended, seq, window)

            TF_list = np.zeros((thre + 2, 2))
            Precision_list = np.ones(thre + 1)
            j = 0

            for i in np.linspace(0, len(score) - 1, thre).astype(int):
                threshold = score_sorted[i]
                pred = score >= threshold
                labels = labels_extended.copy()
                existence = 0

                for seg in L:
                    labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * pred[seg[0]:seg[1] + 1]
                    if (pred[seg[0]:(seg[1] + 1)] > 0).any():
                        existence += 1
                for seg in seq:
                    labels[seg[0]:seg[1] + 1] = 1

                TP = 0
                N_labels = 0
                for seg in l:
                    TP += np.dot(labels[seg[0]:seg[1] + 1], pred[seg[0]:seg[1] + 1])
                    N_labels += np.sum(labels[seg[0]:seg[1] + 1])

                TP += tp[j]
                FP = N_pred[j] - TP

                existence_ratio = existence / len(L)

                P_new = (P + N_labels) / 2
                recall = min(TP / P_new, 1)

                TPR = recall * existence_ratio
                N_new = len(labels) - P_new
                FPR = FP / N_new

                Precision = TP / N_pred[j]

                j += 1
                TF_list[j] = [TPR, FPR]
                Precision_list[j] = Precision

            TF_list[j + 1] = [1, 1]  # otherwise, range-AUC will stop earlier than (1,1)

            tpr_3d[window] = TF_list[:, 0]
            fpr_3d[window] = TF_list[:, 1]
            prec_3d[window] = Precision_list

            width = TF_list[1:, 1] - TF_list[:-1, 1]
            height = (TF_list[1:, 0] + TF_list[:-1, 0]) / 2
            AUC_range = np.dot(width, height)
            auc_3d[window] = (AUC_range)

            # width_PR = TF_list[1:-1, 0] - TF_list[:-2, 0]
            # height_PR = Precision_list[1:]

            # AP_range = np.dot(width_PR, height_PR)
            # ap_3d[window] = AP_range

        return tpr_3d, fpr_3d, prec_3d, window_3d, sum(auc_3d) / len(window_3d), None

    def RangeAUC_volume_opt_mem(self, labels_original, score, windowSize, thre=250):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels_original)
        seq = self.range_convers_new(labels_original)
        l = self.new_sequence(labels_original, seq, windowSize)

        score_sorted = -np.sort(-score)

        tpr_3d = np.zeros((windowSize + 1, thre + 2))
        fpr_3d = np.zeros((windowSize + 1, thre + 2))
        prec_3d = np.zeros((windowSize + 1, thre + 1))

        auc_3d = np.zeros(windowSize + 1)
        ap_3d = np.zeros(windowSize + 1)

        tp = np.zeros(thre)
        N_pred = np.zeros(thre)
        p = np.zeros((thre, len(score)))

        for k, i in enumerate(np.linspace(0, len(score) - 1, thre).astype(int)):
            threshold = score_sorted[i]
            pred = score >= threshold
            p[k] = pred
            N_pred[k] = np.sum(pred)

        for window in window_3d:
            labels_extended = self.sequencing(labels_original, seq, window)
            L = self.new_sequence(labels_extended, seq, window)

            TF_list = np.zeros((thre + 2, 2))
            Precision_list = np.ones(thre + 1)
            j = 0

            for i in np.linspace(0, len(score) - 1, thre).astype(int):
                labels = labels_extended.copy()
                existence = 0

                for seg in L:
                    labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * p[j][seg[0]:seg[1] + 1]
                    if (p[j][seg[0]:(seg[1] + 1)] > 0).any():
                        existence += 1
                for seg in seq:
                    labels[seg[0]:seg[1] + 1] = 1

                N_labels = 0
                TP = 0
                for seg in l:
                    TP += np.dot(labels[seg[0]:seg[1] + 1], p[j][seg[0]:seg[1] + 1])
                    N_labels += np.sum(labels[seg[0]:seg[1] + 1])

                TP += tp[j]
                FP = N_pred[j] - TP

                existence_ratio = existence / len(L)

                P_new = (P + N_labels) / 2
                recall = min(TP / P_new, 1)

                TPR = recall * existence_ratio

                N_new = len(labels) - P_new
                FPR = FP / N_new
                Precision = TP / N_pred[j]
                j += 1

                TF_list[j] = [TPR, FPR]
                Precision_list[j] = Precision

            TF_list[j + 1] = [1, 1]
            tpr_3d[window] = TF_list[:, 0]
            fpr_3d[window] = TF_list[:, 1]
            prec_3d[window] = Precision_list

            width = TF_list[1:, 1] - TF_list[:-1, 1]
            height = (TF_list[1:, 0] + TF_list[:-1, 0]) / 2
            AUC_range = np.dot(width, height)
            auc_3d[window] = (AUC_range)

            width_PR = TF_list[1:-1, 0] - TF_list[:-2, 0]
            height_PR = Precision_list[1:]
            AP_range = np.dot(width_PR, height_PR)
            ap_3d[window] = (AP_range)
        return tpr_3d, fpr_3d, prec_3d, window_3d, sum(auc_3d) / len(window_3d), sum(ap_3d) / len(window_3d)


    def metric_VUS_pred(self, labels, preds, windowSize):
        window_3d = np.arange(0, windowSize + 1, 1)
        P = np.sum(labels)
        seq = self.range_convers_new(labels)
        l = self.new_sequence(labels, seq, windowSize)

        recall_3d = np.zeros((windowSize + 1))
        prec_3d = np.zeros((windowSize + 1))
        f_3d = np.zeros((windowSize + 1))

        N_pred = np.sum(preds)

        for window in window_3d:

            labels_extended = self.sequencing(labels, seq, window)
            L = self.new_sequence(labels_extended, seq, window)
                
            labels = labels_extended.copy()
            existence = 0

            for seg in L:
                labels[seg[0]:seg[1] + 1] = labels_extended[seg[0]:seg[1] + 1] * preds[seg[0]:seg[1] + 1]
                if (preds[seg[0]:(seg[1] + 1)] > 0).any():
                    existence += 1
            for seg in seq:
                labels[seg[0]:seg[1] + 1] = 1

            TP = 0
            N_labels = 0
            for seg in l:
                TP += np.dot(labels[seg[0]:seg[1] + 1], preds[seg[0]:seg[1] + 1])
                N_labels += np.sum(labels[seg[0]:seg[1] + 1])

            P_new = (P + N_labels) / 2
            recall = min(TP / P_new, 1)
            Precision = TP / N_pred            

            recall_3d[window] = recall
            prec_3d[window] = Precision
            f_3d[window] = 2 * Precision * recall / (Precision + recall) if (Precision + recall) > 0 else 0

        return sum(recall_3d) / len(window_3d), sum(prec_3d) / len(window_3d), sum(f_3d) / len(window_3d)
    




if __name__ == "__main__":
    # labels = "0010001111111000011000"
    # labels = np.array([int(i) for i in labels])
    # preds =  "0010000011000001000001"
    # preds = np.array([int(i) for i in preds])

    metricor = basic_metricor()
    # f1,pre,rec = metricor.metric_Reduced_F1(labels, preds, preds)
    # print("Reduced F1:", f1, "Precision:", pre, "Recall:", rec)


    
    theta, alpha, delta, graph = 0.5, 0.8, 600, 'none'  #default values
    assert(0.0 <= theta <= 1.0)
    assert(0.0 <= alpha <= 1.0)
    assert(isinstance(delta, int))
    assert(graph == 'screen' or graph == 'file' or graph == 'none' or graph == 'all')
    import numpy as np
    labels = "0010001111111000011000"
    labels = np.array([int(i) for i in labels])
    preds =  "0010000011000001000001"
    preds = np.array([int(i) for i in preds])

    result = metricor.metric_PA_percentile_K(labels, score=preds)
    print(result)

    # f1, pre, rec = metricor.metric_eTaPR_F1(labels, preds=preds)
    # print(f1, pre, rec)
    # for key, value in result.items():
    #     print(f"{key}: {value}")


"""
This function is adapted from [TimeEval-algorithms] by [CodeLionX&wenig]
Original source: [https://github.com/TimeEval/TimeEval-algorithms]
"""

import numpy as np
from dataclasses import dataclass
from .models.base import BaseDetector
from .utils.utility import zscore

class FFT(BaseDetector):

    def __init__(self, ifft_parameters=5, local_neighbor_window=21, local_outlier_threshold=0.6, max_region_size=50, max_sign_change_distance=10, normalize=True):
        super().__init__()

        self.ifft_parameters = ifft_parameters
        self.local_neighbor_window = local_neighbor_window
        self.local_outlier_threshold = local_outlier_threshold
        self.max_region_size = max_region_size
        self.max_sign_change_distance = max_sign_change_distance
        self.normalize = normalize
        self.decision_scores_ = None

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods."""
        n_samples, n_features = X.shape
        if self.normalize: 
            if n_features == 1:
                X = zscore(X, axis=0, ddof=0)
            else:
                X = zscore(X, axis=1, ddof=1)
        self.data = X
        self.decision_scores_ = self.detect_anomalies()  
        return self

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector."""
        n_samples, n_features = X.shape
        decision_scores_ = np.zeros(n_samples)
        self.data = X
        local_outliers = self.calculate_local_outliers()
        if not local_outliers:
            print("No local outliers detected.")
            return np.zeros_like(self.data)

        regions = self.calculate_region_outliers(local_outliers)
        anomaly_scores = np.zeros_like(self.data)
        for region in regions:
            start_index = local_outliers[region.start_idx].index
            end_index = local_outliers[region.end_idx].index
            anomaly_scores[start_index:end_index + 1] = region.score

        decision_scores_ = anomaly_scores
        return decision_scores_

    @staticmethod
    def reduce_parameters(f: np.ndarray, k: int) -> np.ndarray:
        transformed = f.copy()
        transformed[k:] = 0
        return transformed

    def calculate_local_outliers(self):
        n = len(self.data)
        k = max(min(self.ifft_parameters, n), 1)
        y = self.reduce_parameters(np.fft.fft(self.data), k)
        f2 = np.real(np.fft.ifft(y))

        so = np.abs(f2 - self.data)
        mso = np.mean(so)
        neighbor_c = self.local_neighbor_window // 2

        scores = []
        score_idxs = []
        for i in range(n):
            if so[i] > mso:
                nav = np.mean(self.data[max(i - neighbor_c, 0):min(i + neighbor_c + 1, n)])
                scores.append(self.data[i] - nav)
                score_idxs.append(i)

        if not scores:
            return []

        ms = np.mean(scores)
        sds = np.std(scores) + 1e-6  
        z_scores = (np.array(scores) - ms) / sds

        return [self.LocalOutlier(index=score_idxs[i], z_score=z_scores[i])
                for i in range(len(scores)) if abs(z_scores[i]) > self.local_outlier_threshold]

    def calculate_region_outliers(self, local_outliers):
        def distance(a: int, b: int) -> int:
            return abs(local_outliers[b].index - local_outliers[a].index)

        regions = []
        i = 0
        n_l = len(local_outliers) - 1
        while i < n_l:
            start_idx = i
            while i < n_l and distance(i, i + 1) <= self.max_sign_change_distance:
                i += 1
            end_idx = i
            if end_idx > start_idx:
                score = np.mean([abs(local_outliers[j].z_score) for j in range(start_idx, end_idx + 1)])
                regions.append(self.RegionOutlier(start_idx=start_idx, end_idx=end_idx, score=score))
            i += 1

        return regions

    @dataclass
    class LocalOutlier:
        index: int
        z_score: float

        @property
        def sign(self) -> int:
            return np.sign(self.z_score)

    @dataclass
    class RegionOutlier:
        start_idx: int
        end_idx: int
        score: float

    def detect_anomalies(self):
        """Detect anomalies by combining local and regional outliers."""
        local_outliers = self.calculate_local_outliers()
        if not local_outliers:
            print("No local outliers detected.")
            return np.zeros_like(self.data)

        regions = self.calculate_region_outliers(local_outliers)
        anomaly_scores = np.zeros_like(self.data)
        for region in regions:
            start_index = local_outliers[region.start_idx].index
            end_index = local_outliers[region.end_idx].index
            anomaly_scores[start_index:end_index + 1] = region.score

        return anomaly_scores


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
CCE: Confidence-Consistency Evaluation for Time Series Anomaly Detection

A comprehensive framework for evaluating time series anomaly detection methods
with confidence-consistency metrics.
"""

# __version__ = "0.2.5"
# __author__ = "EmorZz1G"
# __email__ = "csemor@mail.scut.edu.cn"
# __license__ = "MIT"
# __url__ = "https://github.com/EmorZz1G/CCE"

import sys
from pathlib import Path


from . import config
from . import cli

_submodules = ["evaluation", "models", "metrics", "data_utils"]

def __getattr__(name):
    if name in _submodules:
        import importlib
        return importlib.import_module(f".{name}", __name__)
    raise AttributeError(f"module {__name__} has no attribute {name}")

__all__ = _submodules

# 定义from cce import *时导出的内容
__all__ = [
    '__version__',
    '__author__', 
    '__email__',
    '__license__',
    '__url__',
    'config',
    'cli',
    'evaluation',
    'models', 
    'metrics',
    'utils',
    'data_utils',
]


# 自动创建全局配置（保持不变）
def _auto_create_global_config():
    """Automatically create global configuration if it doesn't exist"""
    try:
        from pathlib import Path
        home_config_path = Path.home() / '.cce' / 'config.yaml'
        
        if not home_config_path.exists():
            from .config import create_install_config
            create_install_config()
            print("✅ CCE global configuration auto-created")
            print("💡 Use 'cce config create' in your projects to copy this configuration")
    except Exception:
        # 静默失败，用户可后续手动创建
        pass

# 执行自动配置
_auto_create_global_config()
del _auto_create_global_config



# # 导入平级包（evaluation、models等与cce平级，在src目录下）
# try:
#     import evaluation
#     # 关键：在sys.modules中注册这些模块，使其看起来像是cce的子包
#     sys.modules['cce.evaluation'] = evaluation
    
# except ImportError as e:
#     try:
#         from src import evaluation
#         sys.modules['cce.evaluation'] = evaluation
#     except ImportError as e2:
#         print(f"❌ 备用导入平级包失败: {e2}")
#         # 如果导入失败，设置为None
#         globals()['evaluation'] = None

# try: 
#     import models

#     sys.modules['cce.models'] = models

# except ImportError as e:
#     try:
#         from src import models
#         sys.modules['cce.models'] = models
#     except ImportError as e2:
#         print(f"❌ 备用导入平级包失败: {e2}")
#         # 如果导入失败，设置为None  
#         globals()['models'] = None


# try: 
#     import utils
#     sys.modules['cce.utils'] = utils
# except ImportError as e:
#     try:
#         from src import utils
#         sys.modules['cce.utils'] = utils
#     except ImportError as e2:
#         print(f"❌ 备用导入平级包失败: {e2}")
#         # 如果导入失败，设置为None  
#         globals()['utils'] = None

# try: 
#     import data_utils
#     sys.modules['cce.data_utils'] = data_utils
# except ImportError as e:
#     try:
#         from src import data_utils
#         sys.modules['cce.data_utils'] = data_utils
#     except ImportError as e2:
#         print(f"❌ 备用导入平级包失败: {e2}")
#         # 如果导入失败，设置为None  
#         globals()['data_utils'] = None


# try: 
#     import metrics
#     sys.modules['cce.metrics'] = metrics
# except ImportError as e:
#     try:
#         from src import metrics
#         sys.modules['cce.metrics'] = metrics
#     except ImportError as e2:
#         print(f"❌ 备用导入平级包失败: {e2}")
#         # 如果导入失败，设置为None  
#         globals()['metrics'] = None



"""
This function is adapted from [pyod] by [yzhao062]
Original source: [https://github.com/yzhao062/pyod]
"""

from __future__ import division
from __future__ import print_function
import warnings

import numpy as np

from joblib import Parallel, delayed
from scipy.stats import skew as skew_sp
from sklearn.utils.validation import check_is_fitted
from sklearn.utils import check_array

from .base import BaseDetector
from ..utils.stat_models import column_ecdf
from ..utils.utility import _partition_estimators
from ..utils.utility import zscore

def skew(X, axis=0):
    return np.nan_to_num(skew_sp(X, axis=axis))

def _parallel_ecdf(n_dims, X):
    """Private method to calculate ecdf in parallel.    
    Parameters
    ----------
    n_dims : int
        The number of dimensions of the current input matrix

    X : numpy array
        The subarray for building the ECDF

    Returns
    -------
    U_l_mat : numpy array
        ECDF subarray.

    U_r_mat : numpy array
        ECDF subarray.
    """
    U_l_mat = np.zeros([X.shape[0], n_dims])
    U_r_mat = np.zeros([X.shape[0], n_dims])

    for i in range(n_dims):
        U_l_mat[:, i: i + 1] = column_ecdf(X[:, i: i + 1])
        U_r_mat[:, i: i + 1] = column_ecdf(X[:, i: i + 1] * -1)
    return U_l_mat, U_r_mat

class COPOD(BaseDetector):
    """COPOD class for Copula Based Outlier Detector.
    COPOD is a parameter-free, highly interpretable outlier detection algorithm
    based on empirical copula models.
    See :cite:`li2020copod` for details.

    Parameters
    ----------
    contamination : float in (0., 0.5), optional (default=0.1)
        The amount of contamination of the data set, i.e.
        the proportion of outliers in the data set. Used when fitting to
        define the threshold on the decision function.
        
    n_jobs : optional (default=1)
        The number of jobs to run in parallel for both `fit` and
        `predict`. If -1, then the number of jobs is set to the
        number of cores.

    Attributes
    ----------
    decision_scores_ : numpy array of shape (n_samples,)
        The outlier scores of the training data.
        The higher, the more abnormal. Outliers tend to have higher
        scores. This value is available once the detector is
        fitted.
    threshold_ : float
        The threshold is based on ``contamination``. It is the
        ``n_samples * contamination`` most abnormal samples in
        ``decision_scores_``. The threshold is calculated for generating
        binary outlier labels.
    labels_ : int, either 0 or 1
        The binary labels of the training data. 0 stands for inliers
        and 1 for outliers/anomalies. It is generated by applying
        ``threshold_`` on ``decision_scores_``.
    """

    def __init__(self, contamination=0.1, n_jobs=1, normalize=True):
        super(COPOD, self).__init__(contamination=contamination)

        #TODO: Make it parameterized for n_jobs
        self.n_jobs = n_jobs
        self.normalize = normalize

    def fit(self, X, y=None):
        """Fit detector. y is ignored in unsupervised methods.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The input samples.
        y : Ignored
            Not used, present for API consistency by convention.
        Returns
        -------
        self : object
            Fitted estimator.
        """
        X = check_array(X)
        if self.normalize: X = zscore(X, axis=1, ddof=1)

        self._set_n_classes(y)
        self.decision_scores_ = self.decision_function(X)
        self.X_train = X
        self._process_decision_scores()
        return self

    def decision_function(self, X):
        """Predict raw anomaly score of X using the fitted detector.
         For consistency, outliers are assigned with larger anomaly scores.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        # use multi-thread execution
        if self.n_jobs != 1:
            return self._decision_function_parallel(X)
        if hasattr(self, 'X_train'):
            original_size = X.shape[0]
            X = np.concatenate((self.X_train, X), axis=0)
        self.U_l = -1 * np.log(column_ecdf(X))
        self.U_r = -1 * np.log(column_ecdf(-X))

        skewness = np.sign(skew(X, axis=0))
        self.U_skew = self.U_l * -1 * np.sign(
            skewness - 1) + self.U_r * np.sign(skewness + 1)
        self.O = np.maximum(self.U_skew, np.add(self.U_l, self.U_r) / 2)
        if hasattr(self, 'X_train'):
            decision_scores_ = self.O.sum(axis=1)[-original_size:]
        else:
            decision_scores_ = self.O.sum(axis=1)
        return decision_scores_.ravel()

    def _decision_function_parallel(self, X):
        """Predict raw anomaly score of X using the fitted detector.
         For consistency, outliers are assigned with larger anomaly scores.
        Parameters
        ----------
        X : numpy array of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only
            if they are supported by the base estimator.
        Returns
        -------
        anomaly_scores : numpy array of shape (n_samples,)
            The anomaly score of the input samples.
        """
        if hasattr(self, 'X_train'):
            original_size = X.shape[0]
            X = np.concatenate((self.X_train, X), axis=0)

        n_samples, n_features = X.shape[0], X.shape[1]

        if n_features < 2:
            raise ValueError(
                'n_jobs should not be used on one dimensional dataset')

        if n_features <= self.n_jobs:
            self.n_jobs = n_features
            warnings.warn("n_features <= n_jobs; setting them equal instead.")

        n_jobs, n_dims_list, starts = _partition_estimators(n_features,
                                                            self.n_jobs)

        all_results = Parallel(n_jobs=n_jobs, max_nbytes=None,
                               verbose=True)(
            delayed(_parallel_ecdf)(
                n_dims_list[i],
                X[:, starts[i]:starts[i + 1]],
            )
            for i in range(n_jobs))

        # recover the results
        self.U_l = np.zeros([n_samples, n_features])
        self.U_r = np.zeros([n_samples, n_features])

        for i in range(n_jobs):
            self.U_l[:, starts[i]:starts[i + 1]] = all_results[i][0]
            self.U_r[:, starts[i]:starts[i + 1]] = all_results[i][1]

        self.U_l = -1 * np.log(self.U_l)
        self.U_r = -1 * np.log(self.U_r)

        skewness = np.sign(skew(X, axis=0))
        self.U_skew = self.U_l * -1 * np.sign(
            skewness - 1) + self.U_r * np.sign(skewness + 1)
        self.O = np.maximum(self.U_skew, np.add(self.U_l, self.U_r) / 2)
        if hasattr(self, 'X_train'):
            decision_scores_ = self.O.sum(axis=1)[-original_size:]
        else:
            decision_scores_ = self.O.sum(axis=1)
        return decision_scores_.ravel()